#!/usr/bin/env python
"""
_SimpleCondorPlugin_

"""
import logging
import os
import os.path
import re
import threading
import time
import classad
import htcondor

from Utils import FileTools
from Utils.IteratorTools import grouper
from WMCore.BossAir.Plugins.BasePlugin import BasePlugin
from WMCore.Credential.Proxy import Proxy
from WMCore.DAOFactory import DAOFactory
from WMCore.FwkJobReport.Report import Report
from WMCore.WMInit import getWMBASE
from WMCore.Lexicon import getIterMatchObjectOnRegexp, WMEXCEPTION_REGEXP, CONDOR_LOG_FILTER_REGEXP
from WMCore.Services.TagCollector.TagCollector import TagCollector

def activityToType(jobActivity):
    """
    Function to map a workflow activity to a generic CMS job type.
    :param jobActivity: a workflow activity string
    :return: string which it maps to

    NOTE: this map is based on the Lexicon.activity list
    """
    activityMap = {"reprocessing": "production",
                   "production": "production",
                   "relval": "production",
                   "harvesting": "production",
                   "storeresults": "production",
                   "tier0": "tier0",
                   "t0": "tier0",
                   "integration": "test",
                   "test": "test"}
    return activityMap.get(jobActivity, "unknown")


class SimpleCondorPlugin(BasePlugin):
    """
    _SimpleCondorPlugin_

    Condor plugin for glide-in submissions
    """

    @staticmethod
    def stateMap():
        """
        For a given condor status mapped, return a global state used by the agent
        NOTE: these keys are populated into bl_status table.
        NOTE2: Timeout must be present in all future plugins (New is recommended too)
        """
        stateMap = {'New': 'Pending',
                    'Idle': 'Pending',
                    'Running': 'Running',
                    'Removed': 'Error',
                    'Completed': 'Complete',
                    'Held': 'Error',
                    'TransferOutput': 'Running',
                    'Suspended': 'Error',
                    'Timeout': 'Error',
                    'Unknown': 'Error'}

        return stateMap

    @staticmethod
    def exitCodeMap():
        """
        HTCondor mapping from the numerical value to its english meaning.
        https://htcondor-wiki.cs.wisc.edu/index.cgi/wiki?p=MagicNumbers
        """
        exitCodeMap = {0: "Unknown",
                       1: "Idle",
                       2: "Running",
                       3: "Removed",
                       4: "Completed",
                       5: "Held",
                       6: "TransferOutput",
                       7: "Suspended"}

        return exitCodeMap

    def __init__(self, config):
        BasePlugin.__init__(self, config)

        self.locationDict = {}

        myThread = threading.currentThread()
        daoFactory = DAOFactory(package="WMCore.WMBS", logger=myThread.logger,
                                dbinterface=myThread.dbi)
        self.locationAction = daoFactory(classname="Locations.GetSiteInfo")

        self.packageDir = None

        # if agent is running in a container, Unpacker.py must come from a directory
        # on the host so the condor schedd can see it
        # config.General.workDir should always be bind mounted to the container
        if getattr(config.Agent, "isDocker", False):
            unpackerPath = os.path.join(config.General.workDir + "/Docker/WMRuntime/Unpacker.py")
        else:
            unpackerPath = os.path.join(getWMBASE(), 'src/python/WMCore/WMRuntime/Unpacker.py')

        if os.path.exists(unpackerPath):
            self.unpacker = unpackerPath
        else:
            self.unpacker = os.path.join(getWMBASE(),
                                         'WMCore/WMRuntime/Unpacker.py')

        self.agent = getattr(config.Agent, 'agentName', 'WMAgent')
        self.sandbox = None

        self.scriptFile = config.JobSubmitter.submitScript

        self.defaultTaskPriority = getattr(config.BossAir, 'defaultTaskPriority', 0)
        self.maxTaskPriority = getattr(config.BossAir, 'maxTaskPriority', 1e7)
        self.jobsPerSubmit = getattr(config.JobSubmitter, 'jobsPerSubmit', 200)
        self.extraMem = getattr(config.JobSubmitter, 'extraMemoryPerCore', 500)

        # Required for global pool accounting
        self.acctGroup = getattr(config.BossAir, 'acctGroup', "production")
        self.acctGroupUser = getattr(config.BossAir, 'acctGroupUser', "cmsdataops")
 
        if hasattr(config.BossAir, 'condorRequirementsString'):
            self.reqStr = config.BossAir.condorRequirementsString
        else:
            self.reqStr = None

        # x509 proxy handling
        proxy = Proxy({'logger': myThread.logger})
        self.x509userproxy = proxy.getProxyFilename()

        # These are added now by the condor client
        #self.x509userproxysubject = proxy.getSubject()
        #self.x509userproxyfqan = proxy.getAttributeFromProxy(self.x509userproxy)
        
        self.tc = TagCollector()

        self.useCMSToken = getattr(config.JobSubmitter, 'useOauthToken', False)

        return

    def submit(self, jobs, info=None):
        """
        _submit_

        Submits jobs to the condor queue
        """
        successfulJobs = []
        failedJobs = []

        if len(jobs) == 0:
            # Then was have nothing to do
            return successfulJobs, failedJobs

        schedd = htcondor.Schedd()

        # Submit the jobs
        for jobsReady in grouper(jobs, self.jobsPerSubmit):

            (sub, jobParams) = self.createSubmitRequest(jobsReady)

            logging.debug("Start: Submitting %d jobs using Condor Python Submit", len(jobParams))
            try:
                submitRes = schedd.submit(sub, itemdata=iter(jobParams))
                clusterId = submitRes.cluster()
            except Exception as ex:
                logging.error("SimpleCondorPlugin job submission failed.")
                logging.exception(str(ex))
                logging.error("Moving on the the next batch of jobs and/or cycle....")

                condorErrorReport = Report()
                condorErrorReport.addError("JobSubmit", 61202, "CondorError", str(ex))
                for job in jobsReady:
                    job['fwjr'] = condorErrorReport
                    failedJobs.append(job)
            else:
                logging.debug("Job submission to condor succeeded, clusterId is %s", clusterId)
                for index, job in enumerate(jobsReady):
                    job['gridid'] = "%s.%s" % (clusterId, index)
                    job['status'] = 'Idle'
                    successfulJobs.append(job)

        # We must return a list of jobs successfully submitted and a list of jobs failed
        logging.info("Done submitting jobs for this cycle in SimpleCondorPlugin")
        return successfulJobs, failedJobs

    def track(self, jobs):
        """
        _track_

        Track the jobs while in condor
        This returns a three-way ntuple
        First, the total number of jobs still running
        Second, the jobs that need to be changed
        Third, the jobs that need to be completed
        """
        jobInfo = {}
        changeList = []
        completeList = []
        runningList = []

        # get info about all active and recent jobs
        logging.debug("SimpleCondorPlugin is going to track %s jobs", len(jobs))

        schedd = htcondor.Schedd()

        logging.debug("Start: Retrieving classAds using Condor Python query")
        try:
            jobAds = schedd.query("WMAgent_AgentName == %s" % classad.quote(self.agent),
                                  ['ClusterId', 'ProcId', 'JobStatus', 'MachineAttrGLIDEIN_CMSSite0'])
            for jobAd in jobAds:
                gridId = "%s.%s" % (jobAd['ClusterId'], jobAd['ProcId'])
                jobStatus = SimpleCondorPlugin.exitCodeMap().get(jobAd.get('JobStatus'), 'Unknown')
                location = jobAd.get('MachineAttrGLIDEIN_CMSSite0', None)
                jobInfo[gridId] = (jobStatus, location)
        except Exception as ex:
            logging.error("Query to condor schedd failed in SimpleCondorPlugin.")
            logging.error("Returning empty lists for all job types...")
            logging.exception(ex)
            return runningList, changeList, completeList

        logging.debug("Finished retrieving %d classAds from Condor", len(jobInfo))

        # now go over the jobs and see what we have
        for job in jobs:

            # if the schedd doesn't know a job, consider it complete
            # doing any further checks is not cost effective
            if job['gridid'] not in jobInfo:
                (newStatus, location) = ('Completed', None)
            else:
                (newStatus, location) = jobInfo[job['gridid']]

            # check for status changes
            if newStatus != job['status']:

                # update location info for Idle->Running transition
                if newStatus == 'Running' and job['status'] == 'Idle':
                    if location:
                        job['location'] = location
                        logging.debug("JobAdInfo: Job location for jobid=%i gridid=%s changed to %s", job['jobid'],
                                      job['gridid'], location)

                job['status'] = newStatus
                job['status_time'] = int(time.time())
                logging.debug("JobAdInfo: Job status for jobid=%i gridid=%s changed to %s", job['jobid'], job['gridid'],
                              job['status'])
                changeList.append(job)

            job['globalState'] = SimpleCondorPlugin.stateMap().get(newStatus)

            # stop tracking finished jobs
            if job['globalState'] in ['Complete', 'Error']:
                completeList.append(job)
            else:
                runningList.append(job)

        logging.debug("SimpleCondorPlugin tracking : %i/%i/%i (Executing/Changing/Complete)",
                      len(runningList), len(changeList), len(completeList))

        return runningList, changeList, completeList

    def complete(self, jobs):
        """
        Do any completion work required

        In this case, look for a returned logfile
        """

        for job in jobs:

            if job.get('cache_dir', None) is None or job.get('retry_count', None) is None:
                # Then we can't do anything
                logging.error("Can't find this job's cache_dir or retry count: %s", job)
                continue

            reportName = os.path.join(job['cache_dir'], 'Report.%i.pkl' % job['retry_count'])
            if os.path.isfile(reportName) and os.path.getsize(reportName) > 0:
                # everything in order, move on
                continue
            elif os.path.isdir(reportName):
                # Then something weird has happened. Report error, do nothing
                logging.error("The job report for job with id %s and gridid %s is a directory", job['id'],
                              job['gridid'])
                logging.error("Ignoring this, but this is very strange")
            else:
                logging.error("No job report for job with id %s and gridid %s", job['id'], job['gridid'])

                if os.path.isfile(reportName):
                    os.remove(reportName)

                # create a report from scratch
                condorReport = Report()
                logOutput = 'Could not find jobReport\n'

                if os.path.isdir(job['cache_dir']):
                    condorErr = "condor.%s.err" % job['gridid']
                    condorOut = "condor.%s.out" % job['gridid']
                    condorLog = "condor.%s.log" % job['gridid']
                    exitCode = 99303
                    exitType = "NoJobReport"
                    for condorFile in [condorErr, condorOut, condorLog]:
                        condorFilePath = os.path.join(job['cache_dir'], condorFile)
                        logOutput += "\n========== %s ==========\n" % condorFile
                        if os.path.isfile(condorFilePath):
                            logTail = FileTools.tail(condorFilePath, 50)
                            logOutput += 'Adding end of %s to error message:\n\n' % condorFile
                            logOutput += logTail
                            logOutput += '\n\n'

                            if condorFile == condorLog:
                                # for condor log, search for the information
                                for matchObj in getIterMatchObjectOnRegexp(condorFilePath, CONDOR_LOG_FILTER_REGEXP):
                                    condorReason = matchObj.group("Reason")
                                    if condorReason:
                                        logOutput += condorReason
                                        if "SYSTEM_PERIODIC_REMOVE" in condorReason or "via condor_rm" in condorReason:
                                            exitCode = 99400
                                            exitType = "RemovedByGlideinOrPeriodicExpr"
                                        else:
                                            exitCode = 99401

                                    siteName = matchObj.group("Site")
                                    if siteName:
                                        condorReport.data.siteName = siteName
                                    else:
                                        condorReport.data.siteName = "NoReportedSite"
                            else:
                                for matchObj in getIterMatchObjectOnRegexp(condorFilePath, WMEXCEPTION_REGEXP):
                                    errMsg = matchObj.group('WMException')
                                    if errMsg:
                                        logOutput += "\n\n%s\n" % errMsg

                                    errMsg = matchObj.group('ERROR')
                                    if errMsg:
                                        logOutput += "\n\n%s\n" % errMsg

                    logOutput += '\n\n'
                    condorReport.addError(exitType, exitCode, exitType, logOutput)
                else:
                    msg = "Serious Error in Completing condor job with id %s!\n" % job['id']
                    msg += "Could not find jobCache directory %s\n" % job['cache_dir']
                    msg += "Creating a new cache_dir for failed job report\n"
                    logging.error(msg)
                    os.makedirs(job['cache_dir'])
                    condorReport.addError("NoJobReport", 99304, "NoCacheDir", logOutput)

                condorReport.save(filename=reportName)

                logging.debug("Created failed job report for job with id %s and gridid %s", job['id'], job['gridid'])

        return

    def updateSiteInformation(self, jobs, siteName, excludeSite):
        """
        _updateSiteInformation_

        Allow or disallow jobs to run at a site.
        Called externally by Ops scripts if a site enters or leaves Down, Draining or Aborted.

        Kill job if after removing site from allowed sites it has nowhere to run.

        Parameters:    excludeSite = False when moving to Normal
                       excludeSite = True when moving to Down, Draining or Aborted
        """
        sd = htcondor.Schedd()
        jobIdToKill = []
        jobtokill = []
        origSiteLists = set()

        try:
            jobAds = sd.query('WMAgent_AgentName =?= %s && JobStatus =?= 1' % classad.quote(self.agent),
                              ['WMAgent_JobID', 'DESIRED_Sites', 'ExtDESIRED_Sites'])

            for jobAd in jobAds:
                jobAdId = jobAd.get('WMAgent_JobID')
                desiredSites = jobAd.get('DESIRED_Sites')
                extDesiredSites = jobAd.get('ExtDESIRED_Sites')
                if excludeSite and siteName == desiredSites:
                    jobIdToKill.append(jobAdId)
                else:
                    origSiteLists.add((desiredSites, extDesiredSites))
            logging.info("Set of %d site list condor combinations", len(origSiteLists))
        except Exception as ex:
            msg = "Failed to query condor schedd: %s" % str(ex)
            logging.exception(msg)
            return jobtokill

        with sd.transaction() as dummyTxn:
            for siteStrings in origSiteLists:
                desiredList = set([site.strip() for site in siteStrings[0].split(",")])
                extDesiredList = set([site.strip() for site in siteStrings[1].split(",")])

                if excludeSite and siteName not in desiredList:
                    continue
                elif not excludeSite and (siteName in desiredList or siteName not in extDesiredList):
                    continue
                elif excludeSite:
                    desiredList.remove(siteName)
                    extDesiredList.add(siteName)
                else:  # well, then include
                    desiredList.add(siteName)
                    extDesiredList.remove(siteName)

                # now put it back in the string format expected by condor
                desiredListStr = ",".join(desiredList)
                extDesiredListStr = ",".join(extDesiredList)

                try:
                    sd.edit('DESIRED_Sites =?= %s && ExtDESIRED_Sites =?= %s' % (classad.quote(siteStrings[0]),
                                                                                 classad.quote(siteStrings[1])),
                            "DESIRED_Sites", classad.quote(str(desiredListStr)))
                    sd.edit('DESIRED_Sites =?= %s && ExtDESIRED_Sites =?= %s' % (classad.quote(siteStrings[0]),
                                                                                 classad.quote(siteStrings[1])),
                            "ExtDESIRED_Sites", classad.quote(str(extDesiredListStr)))
                except RuntimeError as ex:
                    msg = 'Failed to condor edit job sites. Could be that no jobs were in condor anymore: %s' % str(ex)
                    logging.warning(msg)

        # now update the list of jobs to be killed
        jobtokill = [job for job in jobs if job['id'] in jobIdToKill]

        return jobtokill

    def kill(self, jobs, raiseEx=False):
        """
        _kill_

        Kill a list of jobs based on the WMBS job names.
        Kill can happen for schedd running on localhost... TBC.
        """
        logging.info("Killing %i jobs from the queue", len(jobs))

        schedd = htcondor.Schedd()
        gridIds = [job['gridid'] for job in jobs]
        try:
            schedd.act(htcondor.JobAction.Remove, gridIds)
        except RuntimeError:
            logging.warning("Error while killing jobs on the schedd: %s", gridIds)
            if raiseEx:
                raise

        return

    def killWorkflowJobs(self, workflow):
        """
        _killWorkflowJobs_

        Kill all the jobs belonging to a specific workflow.
        """
        logging.info("Going to remove all the jobs for workflow %s", workflow)

        schedd = htcondor.Schedd()

        try:
            schedd.act(htcondor.JobAction.Remove, "WMAgent_RequestName == %s" % classad.quote(str(workflow)))
        except RuntimeError:
            logging.warning("Error while killing jobs on the schedd: WMAgent_RequestName=%s", workflow)

        return

    def updateJobInformation(self, workflow, **kwargs):
        """
        _updateJobInformation_

        Update job information for all jobs in the workflow and task,
        the change will take effect if the job is Idle or becomes idle.

        The currently supported changes are only priority for which both the task (taskPriority)
        and workflow priority (requestPriority) must be provided.

        Since the default priority is very high, we only need to adjust new priorities
        for processing/production task types (which have a task priority of 0)
        """
        schedd = htcondor.Schedd()

        if 'requestPriority' in kwargs:
            newPriority = int(kwargs['requestPriority'])
            try:
                constraint = "WMAgent_RequestName =?= %s" % classad.quote(str(workflow))
                constraint += " && JobPrio =!= %d" % newPriority
                constraint += " && stringListMember(CMS_JobType, %s) " % classad.quote(str("Production, Processing"))
                schedd.edit(constraint, 'JobPrio', classad.Literal(newPriority))
            except Exception as ex:
                logging.error("Failed to update JobPrio for WMAgent_RequestName=%s", str(workflow))
                logging.exception(ex)

        return


    def getJobParameters(self, jobList):
        """
        _getJobParameters_

        Return a list of dictionaries with submit parameters per job.
        """

        undefined = 'UNDEFINED'
        jobParameters = []
        # fetch an up-to-date list of CMSSW micro-architectures
        rel_microarchs = self.tc.defaultMicroArchVersionNumberByRelease()

        for job in jobList:
            ad = {}

            ad['initial_Dir'] = job['cache_dir']
            ad['transfer_input_files'] = "%s,%s/%s,%s" % (job['sandbox'], job['packageDir'],
                                                   'JobPackage.pkl', self.unpacker)
            ad['Arguments'] = "%s %i %s" % (os.path.basename(job['sandbox']), job['id'], job["retry_count"])
            ad['transfer_output_files'] = "Report.%i.pkl,wmagentJob.log" % job["retry_count"]

            jobRequirements = []
            # Do not define custom Requirements for Volunteer resources
            if self.reqStr:
                jobRequirements.append(self.reqStr)

            ad['My.x509userproxy'] = classad.quote(self.x509userproxy)

            # Allow oauth based token authentication
            if self.useCMSToken:
                ad['use_oauth_services'] = "cms"

            sites = ','.join(sorted(job.get('possibleSites')))
            ad['My.DESIRED_Sites'] = classad.quote(str(sites))
            sites = ','.join(sorted(job.get('potentialSites')))
            ad['My.ExtDESIRED_Sites'] = classad.quote(str(sites))
            ad['My.CMS_JobRetryCount'] = str(job['retry_count'])
            ad['My.WMAgent_RequestName'] = classad.quote(job['request_name'])
            match = re.compile("^[a-zA-Z0-9_]+_([a-zA-Z0-9]+)-").match(job['request_name'])
            if match:
                ad['My.CMSGroups'] = classad.quote(match.groups()[0])
            else:
                ad['My.CMSGroups'] = undefined
            ad['My.WMAgent_JobID'] = str(job['jobid'])
            ad['My.WMAgent_SubTaskName'] = classad.quote(job['task_name'])
            ad['My.CMS_JobType'] = classad.quote(job['task_type'])
            ad['My.CMS_Type'] = classad.quote(activityToType(job['activity']))
            ad['My.CMS_RequestType'] = classad.quote(job['requestType'])
            if job.get('physicsTaskType', None):
                ad['My.CMS_extendedJobType'] = classad.quote(job['physicsTaskType'])
            else:
                ad['My.CMS_extendedJobType'] = classad.quote(job['task_type'])
            if job.get('campaignName') is None:
                ad['My.CMS_CampaignName'] = undefined
            else:
                ad['My.CMS_CampaignName'] = classad.quote(job['campaignName'])

            # Handling for AWS, cloud and opportunistic resources
            ad['My.AllowOpportunistic'] = str(job.get('allowOpportunistic', False))
            if job.get('inputDataset'):
                ad['My.DESIRED_CMSDataset'] = classad.quote(job['inputDataset'])
            else:
                ad['My.DESIRED_CMSDataset'] = undefined
            if job.get('inputDatasetLocations'):
                sites = ','.join(sorted(job['inputDatasetLocations']))
                ad['My.DESIRED_CMSDataLocations'] = classad.quote(str(sites))
            else:
                ad['My.DESIRED_CMSDataLocations'] = undefined
            if job.get('inputPileup'):
                cmsPileups=','.join(sorted(job['inputPileup']))
                ad['My.DESIRED_CMSPileups'] = classad.quote(str(cmsPileups))
            else:
                ad['My.DESIRED_CMSPileups'] = undefined
            # HighIO
            ad['My.Requestioslots'] = str(1 if job['task_type'] in ["Merge", "Cleanup", "LogCollect"] else 0)
            # GPU resource handling
            # while we do not support a third option for RequiresGPU, make a binary decision
            if job['requiresGPU'] == "required":
                ad['My.RequiresGPU'] = "1"
                ad['request_gpus'] = "1"
            else:
                ad['My.RequiresGPU'] = "0"
                ad['request_gpus'] = "0"
            if job.get('gpuRequirements', None):
                ad['My.DESIRED_GPUMemoryMB'] = str(job['gpuRequirements']['GPUMemoryMB'])
                # CUDACapabilities is a list of strings, with each string matching this regex: r"^\d+.\d$"
                # E.g.: ["1.0", "10.0", "2.1"]
                cudaCapabilities = sorted(job['gpuRequirements']['CUDACapabilities'], key=float)
                
                ad['My.DESIRED_GPUMinimumCapability'] = str(cudaCapabilities[0])
                ad['My.DESIRED_GPUMaximumCapability'] = str(cudaCapabilities[-1])
                ad['My.DESIRED_GPURuntime'] = str(job['gpuRequirements']['CUDARuntime'])
            else:
                ad['My.DESIRED_GPUMemoryMB'] = undefined
                ad['My.DESIRED_GPUMinimumCapability'] = undefined
                ad['My.DESIRED_GPUMaximumCapability'] = undefined
                ad['My.DESIRED_GPURuntime'] = undefined
            if job.get('jobExtraMatchRequirements'):
                ad['My.DESIRED_ExtraMatchRequirements'] = job['jobExtraMatchRequirements']
                jobRequirements.append(job['jobExtraMatchRequirements'])

            # Assign GPU related HTCondor macros classads to DESIRED classads above
            ad['gpus_minimum_memory'] = ad['My.DESIRED_GPUMemoryMB']
            ad['gpus_minimum_capability'] = ad['My.DESIRED_GPUMinimumCapability']
            ad['gpus_maximum_capability'] = ad['My.DESIRED_GPUMaximumCapability']
            ad['gpus_minimum_runtime'] = ad['My.DESIRED_GPURuntime']

            # Performance and resource estimates (including JDL magic tweaks)
            origCores = job.get('numberOfCores', 1)
            estimatedMins = int(job['estimatedJobTime'] / 60.0) if job.get('estimatedJobTime') else 12 * 60
            estimatedMinsSingleCore = estimatedMins * origCores
            # For now, assume a 15 minute job startup overhead -- condor will round this up further
            ad['My.EstimatedSingleCoreMins'] = str(estimatedMinsSingleCore)
            ad['My.OriginalMaxWallTimeMins'] = str(estimatedMins)
            ad['My.MaxWallTimeMins'] = 'WMCore_ResizeJob ? (EstimatedSingleCoreMins/RequestCpus + 15) : OriginalMaxWallTimeMins'
            requestMemory = int(job['estimatedMemoryUsage']) if job.get('estimatedMemoryUsage', None) else 1000
            ad['My.OriginalMemory'] = str(requestMemory)
            ad['My.ExtraMemory'] = str(self.extraMem)
            ad['request_memory'] = 'OriginalMemory + ExtraMemory * (WMCore_ResizeJob ? (RequestCpus-OriginalCpus) : 0)'
            requestDisk = int(job['estimatedDiskUsage']) if job.get('estimatedDiskUsage', None) else 20 * 1000 * 1000 * origCores
            ad['request_disk'] = str(requestDisk)
            # Set up JDL for multithreaded jobs.
            # By default, RequestCpus will evaluate to whatever CPU request was in the workflow.
            # If the job is labelled as resizable, then the logic is more complex:
            # - If the job is running in a slot with N cores, this should evaluate to N
            # - If the job is being matched against a machine, match all available CPUs, provided
            # they are between min and max CPUs.
            # - Otherwise, just use the original CPU count.
            ad['My.MinCores'] = str(job.get('minCores', max(1, origCores / 2)))
            ad['My.MaxCores'] = str(max(int(job.get('maxCores', origCores)), origCores))
            ad['My.OriginalCpus'] = str(origCores)
            # Prefer slots that are closest to our MaxCores without going over.
            # If the slot size is _greater_ than our MaxCores, we prefer not to
            # use it - we might unnecessarily fragment the slot.
            ad['Rank'] = 'isUndefined(Cpus) ? 0 : ifThenElse(Cpus > MaxCores, -Cpus, Cpus)'
            # Record the number of CPUs utilized at match time.  We'll use this later
            # for monitoring and accounting.  Defaults to 0; once matched, it'll
            # put an attribute in the job  MATCH_EXP_JOB_GLIDEIN_Cpus = 4
            ad['My.JOB_GLIDEIN_Cpus'] = classad.quote("$$(Cpus:0)")
            # Make sure the resize request stays within MinCores and MaxCores.
            ad['My.RequestResizedCpus'] = '(Cpus>MaxCores) ? MaxCores : ((Cpus < MinCores) ? MinCores : Cpus)'
            # If the job is running, then we should report the matched CPUs in RequestCpus - but only if there are sane
            # values.  Otherwise, we just report the original CPU request
            ad['My.JobCpus'] = ('((JobStatus =!= 1) && (JobStatus =!= 5) && !isUndefined(MATCH_EXP_JOB_GLIDEIN_Cpus) '
                              '&& (int(MATCH_EXP_JOB_GLIDEIN_Cpus) isnt error)) ? int(MATCH_EXP_JOB_GLIDEIN_Cpus) : OriginalCpus')
            # Cpus is taken from the machine ad - hence it is only defined when we are doing negotiation.
            # Otherwise, we use either the cores in the running job (if available) or the original cores.
            ad['request_cpus'] = 'WMCore_ResizeJob ? (!isUndefined(Cpus) ? RequestResizedCpus : JobCpus) : OriginalCpus'
            ad['My.WMCore_ResizeJob'] = str(job.get('resizeJob', False))
            taskPriority = int(job.get('taskPriority', 1))
            priority = int(job.get('wf_priority', 0))
            ad['My.JobPrio'] = str(int(priority + taskPriority * self.maxTaskPriority))
            ad['My.PostJobPrio1'] = str(int(-1 * len(job.get('potentialSites', []))))
            ad['My.PostJobPrio2'] = str(int(-1 * job['task_id']))
            # Add OS requirements for jobs
            requiredOSes = self.scramArchtoRequiredOS(job.get('scramArch'))
            ad['My.REQUIRED_OS'] = classad.quote(requiredOSes)
            cmsswVersions = ','.join(job.get('swVersion'))
            ad['My.CMSSW_Versions'] = classad.quote(cmsswVersions)
            requiredArchs = self.scramArchtoRequiredArch(job.get('scramArch'))
            if not requiredArchs:  # only Cleanup jobs should not have ScramArch defined
                ad['My.REQUIRED_ARCH'] = undefined
                jobRequirements.append("(TARGET.Arch =!= REQUIRED_ARCH)")
            else:
                ad['My.REQUIRED_ARCH'] = classad.quote(str(requiredArchs))
                jobRequirements.append("stringListMember(TARGET.Arch, REQUIRED_ARCH)")

            # Inject a microarchitecture classad. If x86_64 not on arch list, return 0
            if 'X86_64' not in requiredArchs.split(","):
                ad['My.REQUIRED_MINIMUM_MICROARCH'] = "0"
            else:
                minMicroArch = self.tc.getGreaterMicroarchVersionNumber(cmsswVersions, rel_microarchs=rel_microarchs)
                ad['My.REQUIRED_MINIMUM_MICROARCH'] = str(minMicroArch) 

            # Now AND all the new job requirements together - else default to empty string
            ad['Requirements'] = " && ".join(jobRequirements)

            jobParameters.append(ad)
             
        return jobParameters


    def createSubmitRequest(self, jobList):
        """
        _createSubmitRequest_

        Return the submit object to pass to htcondor.Submit()

        """

        sub = htcondor.Submit("""
            universe = vanilla
            should_transfer_files = YES
            when_to_transfer_output = ON_EXIT
            notification = NEVER
            log_xml = True
            rank = 0.0
            transfer_input = False
            Output = condor.$(Cluster).$(Process).out
            Error = condor.$(Cluster).$(Process).err
            Log = condor.$(Cluster).$(Process).log
            +JobLeaseDuration = (isUndefined(MachineAttrMaxHibernateTime0) ? 1200 : MachineAttrMaxHibernateTime0)
            +PeriodicRemove = ( JobStatus =?= 5 ) && ( time() - EnteredCurrentStatus > 10 * 60 )
            +PeriodicRemoveReason = PeriodicRemove ? "Job automatically removed for being in Held status" : ""
            """)
        sub['executable'] = self.scriptFile

        # Required for global pool accounting
        sub['My.WMAgent_AgentName'] = classad.quote(self.agent)
        sub['accounting_group'] = self.acctGroup
        sub['accounting_group_user'] = self.acctGroupUser
        sub['My.JobMachineAttrs'] = classad.quote("GLIDEIN_CMSSite,GLIDEIN_Gatekeeper")
        sub['My.JobAdInformationAttrs'] = classad.quote("JobStatus,QDate,EnteredCurrentStatus,JobStartDate,DESIRED_Sites,ExtDESIRED_Sites,WMAgent_JobID,MachineAttrGLIDEIN_CMSSite0,MyType")

        # entries required for monitoring
        sub['My.CMS_WMTool'] = classad.quote("WMAgent")
        sub['My.CMS_SubmissionTool'] = classad.quote("WMAgent")

        jobParameters = self.getJobParameters(jobList)

        return sub, jobParameters



#!/usr/bin/env python
"""
This module provides helper functions to obtain and handle CouchDB Replication data:
    - getSchedulerJobDocs get replication status based on scheduler information
    - getReplicatorDocs get replication status based on replicator information
    - compareReplicationStatus compares previous and current statuses
    - formatPrometheusMetrics format status metrics in Prometheus format
    - createAlerts create alerts from given status dict
    - checkStatus perform all checks for couchdb replication

Example of using Flask framework to serve prometheus metrics about CouchDB replication

import requests
from flask import Flask, Response
import threading
import time

app = Flask(__name__)
status_cache = {}

@app.route("/metrics")
def metrics():
    return Response(formatPrometheusMetrics(status_cache), mimetype="text/plain")

def daemonCouchReplicationStatus(interval=30):
    global status_cache
    while True:
        new_status = getSchedulerJobDocs(COUCHDB_URL, USERNAME, PASSWORD)
        status_cache = new_status
        time.sleep(interval)

if __name__ == "__main__":
    # Start the background thread to update replication status periodically
    threading.Thread(target=daemonCouchReplicationStatus, daemon=True).start()
    # Run the Flask app
    app.run(host="0.0.0.0", port=8000)
"""

import os
import json
import requests
import tempfile

# WMCore modules
from Utils.CertTools import cert, ckey, caBundle


def getSchedulerJobDocs(couchdbUrl):
    """
    Fetch CouchDB replication statuses. The logic is based on /_scheduler/jobs CouchDB end-point
    see https://docs.couchdb.org/en/stable/api/server/common.html#api-server-scheduler-jobs
    :param couchdbUrl: url of couch db
    :return: dictionary of statuses for all found replication documents
    """
    username, password = couchCredentials()
    auth = (username, password) if username and password else None
    try:
        response = requests.get(f"{couchdbUrl}/_scheduler/jobs", auth=auth)
        response.raise_for_status()
        data = response.json()

        statuses = {}
        for job in data.get('jobs', []):
            doc_id = job.get('doc_id') or job.get('id')
            source = job.get('source')
            target = job.get('target')
            history = job.get('history', [])
            info = job.get('info', {})

            # Determine current state from latest history item
            state = history[0]['type'] if history else 'unknown'

            # Detect error if 'crashed' exists in any history entry
            error = None
            for h in history:
                if h.get('type') == 'crashed':
                    error = f"Job previous crashed at {h.get('timestamp')} due to {h.get('reason')}"
                    break

            # check info document
            if info and info.get('doc_write_failures', 0) != 0:
                error = f"found failure of replication jobs in {couchdbUrl}/_scheduler/jobs "
                state = "error"
                # try to get more info about the error
                try:
                    response = requests.get(f"{couchdbUrl}/_scheduler/docs/_replicator/{doc_id}", auth=auth)
                    response.raise_for_status()
                    data = response.json()
                    error += f" Replicator state for {doc_id}: "
                    error += json.dumps(data)
                except:
                    pass

            statuses[doc_id] = {
                'state': state,
                'source': source,
                'target': target,
                'error': error,
                'history': history
                }

        return statuses
    except requests.RequestException as e:
        print(f"Error fetching scheduler jobs: {e}")
        return {}


def getReplicatorDocs(url=None):
    """
    Helper function to get all replicator docs and return summary dictionary
    :param url: url of the couchdb
    :return: replication summary dictionary
    """
    username, password = couchCredentials()
    auth = (username, password) if username and password else None
    if not url:
        url = "http://localhost:5984"
    headers = {"Accept": "application/json"}

    # Get list of all documents in _replicator
    r = requests.get(f"{url}/_replicator/_all_docs?include_docs=true",
                     headers=headers, auth=auth)
 
    if r.status_code != 200:
        raise Exception(f"Failed to fetch replication docs: {r.text}")

    data = r.json()
    result = {}

    for row in data.get("rows", []):
        doc = row.get("doc", {})
        doc_id = doc.get("_id")
        if doc_id.startswith("_design/"):
            continue  # skip design docs

        summary = {
            "state": doc.get("_replication_state"),
            "source": doc.get("source"),
            "target": doc.get("target"),
            "error": doc.get("_replication_state_reason"),
            "history": []
        }

        history = doc.get("_replication_history", [])
        for h in history:
            entry = {
                "timestamp": h.get("start_time") or h.get("end_time"),
                "type": h.get("type") or "unknown"
            }
            summary["history"].append(entry)

        result[doc_id] = summary

    return result


def compareReplicationStatus(prev, curr):
    """
    Helper function to compare replication status from previous to current state
    :param prev: previous replication status dictionary
    :param curr: current replication status dictionary
    :return: dictionary of changes
    """
    changes = {}
    for key in curr:
        if key not in prev or prev[key] != curr[key]:
            changes[key] = {
                    'old': prev.get(key),
                    'new': curr[key]
                    }
    return changes


def formatPrometheusMetrics(statuses):
    """
    Helper function to provide Prometheus metrics from given status dictionary
    :param statuses: replication status dictionary
    :return: prometheus metrics
    """
    states = {'error': -1, 'completed': 0, 'started': 1, 'added': 2, 'waiting': 3, 'triggered': 4, 'failed': 5}
    lines = [
            f'# HELP couchdb_replication_state Replication state: {states}',
            '# TYPE couchdb_replication_state gauge'
            ]
    for key, status in statuses.items():
        label = f'replId="{key}",source="{status["source"]}",target="{status["target"]}"'
        value = 0   # default error/other
        for k, v in states.items():
            if status['state'] == k:
                value = v
                break
        lines.append(f'couchdb_replication_state{{{label}}} {value}')
    return '\n'.join(lines)


def createAlerts(statuses):
    """
    Helper function to check alerts of replication status dictionary
    :param statuses: replication status dictionary
    :return: alerts dictionary
    """
    alerts = {}
    for key, status in statuses.items():
        if status['state'] != 'completed':
            alerts[key] = f"Replication state for {key} is '{status['state']}', error: {status['error']}"
    return alerts


def couchCredentials():
    """
    Select CouchDB credentials from provided secrets file
    :return: tuple of (user, password)
    """
    fname = os.getenv('WMAGENT_SECRETS_LOCATION', '')
    if fname == "":
        raise Exception("No WMAGENT_SECRETS_LOCATION in environment")
    user = ''
    password = ''
    data = ''
    with open(fname, 'r', encoding="utf-8") as istream:
        data = istream.read()
    for item in data.split('\n'):
        if 'COUCH_USER' in item:
            user = item.split('=')[-1]
        if 'COUCH_PASS' in item:
            password = item.split('=')[-1]
    return user, password


def checkStatus(url=None, prevStatus=None, kind="scheduler"):
    """
    Perform check of replication statuses
    :param url: couchdb URL
    :param prevStatus: previous status dictionary
    :param kind: kind of data look-up, e.g. scheduler or replicator
    :return: dictionary of current couchdb replication

    Here is an example of such dictionary structure:
    {'current_status': currStatus (dictionary),
     'previous_status': prevStatus (dictionary),
     'changes': changes (dictionary),
     'metrics': metrics (string),
     'alerts': alerts (dictionary)}

    Then, current and previous status dictionaries have the following form:
    {
      "14843c24643f8960eb159f5912f0f938": {
        "state": "started",
        "source": "https://xxx.cern.ch/couchdb/workqueue/",
        "target": "http://127.0.0.1:5984/workqueue_inbox/",
        "error": "Job previously crashed at 2025-05-05T18:47:11Z due to {changes_reader_died,{timeout,ibrowse_stream_cleanup}}",
        "history": [
          {
            "timestamp": "2025-05-05T18:47:11Z",
            "type": "started"
          },
          ...
         ]
      },
      "14843c24643f8960eb159f5912f0e51e": {
        "state": "started",
        "source": "http://127.0.0.1:5984/wmagent_summary/",
        "target": "https://xxx.cern.ch/couchdb/wmstats/",
        "error": null,
        "history": [
          {
            "timestamp": "2025-04-09T11:19:36Z",
            "type": "started"
          },
          {
            "timestamp": "2025-04-09T11:19:36Z",
            "type": "added"
          }
        ]
      },
    ...
    }
    """
    if not prevStatus:
        prevStatus = {}
    if not url:
        url = "http://localhost:5984"

    # first let's get statuses of documents
    if kind == "scheduler":
        currStatus = getSchedulerJobDocs(url)
    elif kind == "replicator":
        currStatus = getReplicatorDocs(url)
    else:
        raise Exception("Unsupported kind of documents '{kind}', should be either scheduler or replicator")

    # now we can find out changes from previous statuses
    changes = compareReplicationStatus(prevStatus, currStatus)

    # construct prometheus metrics with current statuses
    metrics = formatPrometheusMetrics(currStatus)

    # construct alerts with current statuses
    alerts = createAlerts(currStatus)

    # build final dictionary to return upstream
    sdict = {'current_status': currStatus,
             'previous_status': prevStatus,
             'changes': changes,
             'metrics': metrics,
             'alerts': alerts}
    return sdict


def getDocCount(url, auth, certTuple, caCert):
    """
    helper function to get document counts
    :param url: url of the couchdb
    :param auth: couchdb authentication credentials tuple
    :param caCert: ca bundle file name
    :return: document count
    """
    resp = requests.get(url, auth=auth, cert=certTuple, verify=caCert or True)
    resp.raise_for_status()
    return resp.json().get('doc_count', -1)


def getReplicationState(url, auth, certTuple, caCert):
    """
    helper function to get replication state from given couchdb url
    :param url: url of the couchdb
    :param auth: couchdb authentication credentials tuple
    :param caCert: ca bundle file name
    :return: tuple of replication state and its time
    """
    resp = requests.get(url, auth=auth, cert=certTuple, verify=caCert or True)
    resp.raise_for_status()
    doc = resp.json()
    return doc.get('_replication_state'), doc.get('_replication_state_time')


def compareCouchInstances(sourceUrl, targetUrl, replUrl):
    """
    Compare the number of documents between source and destination CouchDB databases.
    Monitor replication if the counts differ but replication status is OK.
 
    Parameters:
    :param sourceUrl: str, e.g. http://localhost:5984/source_db
    :param targetUrl: str, e.g. http://localhost:5984/dest_db
    :param replUrl: str, e.g. http://localhost:5984/_replicator/<replId>
    """
    user, password = couchCredentials()
    auth = (user, password)
    sdict = {}
    userCert = cert() if cert() else ''
    userCkey = ckey() if ckey() else ''
    if userCkey == '' or userCert == '':
        return sdict
    certTuple = (userCert, userCkey)
    with tempfile.NamedTemporaryFile(mode='w+', suffix=".pem", delete=True) as tfile:
        capath = os.environ.get("X509_CERT_DIR", '/etc/grid-security/certificates')
        cacerts = caBundle(capath)
        tfile.write(cacerts)
        tfile.flush()

        sourceCount = getDocCount(sourceUrl, auth, certTuple, tfile.name)
        targetCount = getDocCount(targetUrl, auth, certTuple, tfile.name)
        state, stateTime = getReplicationState(replUrl, auth, certTuple, tfile.name)

        sdict = {
                "source": sourceUrl,
                "target": targetUrl,
                "source_count": sourceCount,
                "target_count": targetCount,
                "state": state,
                "state_timestamp": stateTime
                }
    return sdict


def exampleReplicationStatus(sourceUrl=None):
    """
    Example function to test replication status either based on scheduler or replicator info
    This function should run on a node with local CouchDB access as all of its logic
    relies on using localhost:5984 URL
    """

    try:
        print(f"checking {sourceUrl}")

        # let's first test scheduler info
        sdict = checkStatus(url=sourceUrl, kind="scheduler")
        print('--- status based on scheduler info ---')
        print(sdict['current_status'])
        print('--- metrics ---')
        print(sdict['metrics'])
        if sdict.get('alerts', None):
            print('--- alerts ---')
            for k, msg in sdict['alerts'].items():
                print(f"{k}: {msg}")

        print()

        # now let's test replicator info
        rdict = checkStatus(url=sourceUrl, kind="replicator")
        print('--- status based on replicator info ---')
        print(rdict['current_status'])
        print('--- metrics ---')
        print(rdict['metrics'])
        if rdict.get('alerts', None):
            print('--- alerts ---')
            for k, msg in rdict['alerts'].items():
                print(f"{k}: {msg}")

    except Exception as exp:
        print(str(exp))


def exampleIndividualDocument(sourceUrl, targetUrl, replUrl):
    """
    Example function how to test check status of particular replication document
    This function should run through CMSWEB frontend URLs as we need to compare
    documents in both source and target CouchDB instances
    :param sourceUrl: source couchdb URL, e.g. https://xxx.cern.ch/couchdb/test_db
    :param targetUrl: target couchdb URL, e.g. https://xxx.cern.ch/couchdb/test_db
    :param replUrl: replication URL, e.g. https://xxx.cern.ch/couchdb/test_db/_replicator/bla
    """
    try:
        result = compareCouchInstances(sourceUrl, targetUrl, replUrl)
        print('--- compare CouchDB Instances ---')
        print('source: ', sourceUrl)
        print('target: ', targetUrl)
        print(result)
    except:
        pass


def test():
    """
    test functions
    """
    import sys
    if len(sys.argv) > 1:
        sourceUrl = sys.argv[1]
        exampleReplicationStatus(sourceUrl)
    else:
        print("Cannot run tests, please provide at least CouchDB source URL, or <srcUrl> <targetUrl> <replicationId>")
    if len(sys.argv) == 4:
        sourceUrl = sys.argv[1]
        targetUrl = sys.argv[2]
        replUrl = sys.argv[3]
        exampleIndividualDocument(sourceUrl, targetUrl, replUrl)

if __name__ == '__main__':
    test()



#!/usr/bin/env python
"""
_ListDataset_

MySQL implementation of DBSBuffer.GetBlockFromDataset
"""




from WMCore.Database.DBFormatter import DBFormatter

class GetBlockFromDataset(DBFormatter):
    """
    _GetBlockFromDataset_

    If given dataset name, get blockname
    """
    sql = """SELECT DISTINCT blockname FROM dbsbuffer_block
               INNER JOIN dbsbuffer_file ON dbsbuffer_block.id = dbsbuffer_file.block_id
               INNER JOIN dbsbuffer_algo_dataset_assoc ON dbsbuffer_file.dataset_algo = dbsbuffer_algo_dataset_assoc.id
               INNER JOIN dbsbuffer_dataset ON dbsbuffer_algo_dataset_assoc.dataset_id = dbsbuffer_dataset.id
               WHERE dbsbuffer_dataset.Path = :dataset
    """

    def execute(self, dataset = None, conn = None,
                transaction = False):
        """
        _execute_

        Either the dataset's ID or path must be specified.
        """

        binds = {'dataset': dataset}

        result = self.dbi.processData(self.sql, binds, conn = conn,
                                          transaction = transaction)

        return self.formatDict(result)



#!/usr/bin/env python
"""
_New_

Oracle implementation of Jobs.New
"""

__all__ = []



from WMCore.WMBS.MySQL.Jobs.New import New as NewJobMySQL

class New(NewJobMySQL):
    sql = """INSERT INTO wmbs_job (jobgroup, name, state, state_time,
                                    couch_record, cache_dir, location, outcome, fwjr_path) VALUES
               (:jobgroup, :name,
                (SELECT id FROM wmbs_job_state WHERE name = 'new'),
                :state_time, :couch_record, :cache_dir,
                (SELECT id FROM wmbs_location WHERE site_name = :location),
                :outcome, :fwjr_path)"""



"""
_HartbeatAPI_

A simple object representing a file in WMBS.
"""

import os

from WMCore.Database.DBExceptionHandler import db_exception_handler
from WMCore.WMConnectionBase import WMConnectionBase


class HeartbeatAPI(WMConnectionBase):
    """
    Generic methods used by all of the WMBS classes.
    """

    def __init__(self, componentName, pollInterval=None, heartbeatTimeout=7200,
                 logger=None, dbi=None):
        """
        ___init___

        Initialize all the database connection attributes and the logging
        attributes.
        Every worker has a different instance of this class, thus they can have
        a different polling interval and a different heartbeat timeout.
        Finally, check to see if a transaction object has been created.
        If none exists, create one but leave the transaction closed.
        """
        self.componentName = componentName
        self.pollInterval = pollInterval
        self.heartbeatTimeout = heartbeatTimeout or 7200
        self.compId = os.getpid()

        WMConnectionBase.__init__(self, daoPackage="WMCore.Agent.Database",
                                  logger=logger, dbi=dbi)

        self.insertComp = self.daofactory(classname="InsertComponent")
        self.existWorker = self.daofactory(classname="ExistWorker")
        self.insertWorker = self.daofactory(classname="InsertWorker")
        self.updateWorker = self.daofactory(classname="UpdateWorker")
        self.updateErrorWorker = self.daofactory(classname="UpdateWorkerError")
        self.getHeartbeat = self.daofactory(classname="GetHeartbeatInfo")
        self.getAllHeartbeat = self.daofactory(classname="GetAllHeartbeatInfo")

    def registerComponent(self):
        """
        Deletes any leftover for a component with the same name and then
        inserts it again with a new PID into wm_components table
        """
        self.insertComp.execute(self.componentName, self.compId, self.heartbeatTimeout,
                                conn=self.getDBConn(), transaction=self.existingTransaction())

    def registerWorker(self, workerName, state="Start"):
        """
        Inserts a worker thread into the database, setting an initial state
        and a polling cycle.
        """
        self.insertWorker.execute(self.componentName, workerName, state,
                                  self.compId, self.pollInterval, cycleTime=0,
                                  conn=self.getDBConn(),
                                  transaction=self.existingTransaction())

    def getComponentID(self, workerName):
        """Retrieve the component_id from the workers table"""

        componentID = self.existWorker.execute(self.componentName, workerName,
                                               conn=self.getDBConn(),
                                               transaction=self.existingTransaction())
        return componentID

    def updateWorkerHeartbeat(self, workerName, state):
        """
        Update a worker's heartbeat and its state
        """
        try:
            self.updateWorker.execute(workerName, state, conn=self.getDBConn(),
                                      transaction=self.existingTransaction())
        except Exception as ex:
            self.logger.warning("Heartbeat update failed! Wait for the next time...:\n%s", str(ex))

    @db_exception_handler
    def updateWorkerCycle(self, workerName, timeSpent, results):
        """
        Update a worker's heartbeat as well as the time spent on that
        cycle and any results returned.
        """
        self.updateWorker.execute(workerName, "Running", timeSpent, results,
                                  conn=self.getDBConn(),
                                  transaction=self.existingTransaction())

    @db_exception_handler
    def updateWorkerError(self, workerName, errorMessage):

        self.updateErrorWorker.execute(self.componentName, workerName, errorMessage,
                                       conn=self.getDBConn(),
                                       transaction=self.existingTransaction())

    def getHeartbeatInfo(self):

        results = self.getHeartbeat.execute(self.componentName, conn=self.getDBConn(),
                                            transaction=self.existingTransaction())

        return results

    def getAllHeartbeatInfo(self):

        results = self.getAllHeartbeat.execute(conn=self.getDBConn(),
                                               transaction=self.existingTransaction())

        return results



"""
_Express_

Express workflow

express processing -> FEVT/RAW/RECO/whatever -> express merge
                      (supports multiple output with different primary datasets)
                   -> ALCARECO -> alca skimming / merging -> ALCARECO
                                                          -> ALCAPROMPT -> end of run alca harvesting -> sqlite -> dropbox upload
                   -> DQM -> merge -> DQM -> periodic dqm harvesting
                                          -> end of run dqm harvesting
"""

from __future__ import division

from future.utils import viewitems

import WMCore.WMSpec.Steps.StepFactory as StepFactory
from Utils.Utilities import makeList, makeNonEmptyList
from WMCore.Lexicon import procstringT0
from WMCore.ReqMgr.Tools.cms import releases, architectures
from WMCore.WMSpec.StdSpecs.StdBase import StdBase


class ExpressWorkloadFactory(StdBase):
    """
    _ExpressWorkloadFactory_

    Stamp out Express workflows.
    """

    def __init__(self):
        StdBase.__init__(self)

        self.inputPrimaryDataset = None
        self.inputProcessedDataset = None

        return

    def buildWorkload(self):
        """
        _buildWorkload_

        Build the workload given all of the input parameters.

        Not that there will be LogCollect tasks created for each processing
        task and Cleanup tasks created for each merge task.

        """
        workload = self.createWorkload()
        workload.setDashboardActivity("t0")

        cmsswStepType = "CMSSW"
        taskType = "Processing"

        # complete output configuration
        for output in self.outputs:
            output['moduleLabel'] = "write_%s_%s" % (output['primaryDataset'],
                                                     output['dataTier'])

        # finalize splitting parameters
        mySplitArgs = self.expressSplitArgs.copy()
        mySplitArgs['algo_package'] = "T0.JobSplitting"

        expressTask = workload.newTask("Express")

        #
        # need to split this up into two separate code paths
        # one is direct reco from the streamer files
        # the other is conversion and then reco
        #
        if self.recoFrameworkVersion is None or self.recoFrameworkVersion == self.frameworkVersion:

            expressRecoStepName = "cmsRun1"

            scenarioArgs = {'globalTag': self.globalTag,
                            'skims': self.alcaSkims,
                            'dqmSeq': self.dqmSequences,
                            'outputs': self.outputs,
                            'inputSource': "DAT"}

            if self.globalTagConnect:
                scenarioArgs['globalTagConnect'] = self.globalTagConnect

            expressOutMods = self.setupProcessingTask(expressTask, taskType,
                                                      scenarioName=self.procScenario,
                                                      scenarioFunc="expressProcessing",
                                                      scenarioArgs=scenarioArgs,
                                                      splitAlgo="Express",
                                                      splitArgs=mySplitArgs,
                                                      stepType=cmsswStepType,
                                                      forceUnmerged=True)
        else:

            expressRecoStepName = "cmsRun2"

            conversionOutMods = self.setupProcessingTask(expressTask, taskType,
                                                         scenarioName=self.procScenario,
                                                         scenarioFunc="repack",
                                                         scenarioArgs={'outputs': [{'dataTier': "RAW",
                                                                                    'eventContent': "ALL",
                                                                                    'primaryDataset': self.specialDataset,
                                                                                    'moduleLabel': "write_RAW"}]},
                                                         splitAlgo="Express",
                                                         splitArgs=mySplitArgs,
                                                         stepType=cmsswStepType,
                                                         forceUnmerged=True)

            # there is only one
            conversionOutLabel = next(iter(conversionOutMods))

            # everything coming after should use the reco CMSSW version and Scram Arch
            self.frameworkVersion = self.recoFrameworkVersion
            self.scramArch = self.recoScramArch

            # add a second step doing the reconstruction
            parentCmsswStep = expressTask.getStep("cmsRun1")
            parentCmsswStepHelper = parentCmsswStep.getTypeHelper()
            parentCmsswStepHelper.keepOutput(False)
            stepTwoCmssw = parentCmsswStep.addTopStep("cmsRun2")
            stepTwoCmssw.setStepType(cmsswStepType)

            template = StepFactory.getStepTemplate(cmsswStepType)
            template(stepTwoCmssw.data)

            stepTwoCmsswHelper = stepTwoCmssw.getTypeHelper()

            stepTwoCmsswHelper.setNumberOfCores(self.multicore)

            stepTwoCmsswHelper.setGlobalTag(self.globalTag)
            stepTwoCmsswHelper.setupChainedProcessing("cmsRun1", conversionOutLabel)
            stepTwoCmsswHelper.cmsswSetup(self.frameworkVersion, softwareEnvironment="",
                                          scramArch=self.scramArch)

            scenarioFunc = "expressProcessing"
            scenarioArgs = {'globalTag': self.globalTag,
                            'skims': self.alcaSkims,
                            'dqmSeq': self.dqmSequences,
                            'outputs': self.outputs,
                            'inputSource': "EDM"}

            if self.globalTagConnect:
                scenarioArgs['globalTagConnect'] = self.globalTagConnect

            configOutput = self.determineOutputModules(scenarioFunc, scenarioArgs)

            expressOutMods = {}
            for outputModuleName in configOutput:
                outputModule = self.addOutputModule(expressTask,
                                                    outputModuleName,
                                                    configOutput[outputModuleName]['primaryDataset'],
                                                    configOutput[outputModuleName]['dataTier'],
                                                    configOutput[outputModuleName].get('filterName', None),
                                                    stepName="cmsRun2", forceUnmerged=True)
                expressOutMods[outputModuleName] = outputModule

            if 'outputs' in scenarioArgs:
                for output in scenarioArgs['outputs']:
                    if 'primaryDataset' in output:
                        del output['primaryDataset']
            if 'primaryDataset' in scenarioArgs:
                del scenarioArgs['primaryDataset']

            stepTwoCmsswHelper.setDataProcessingConfig(self.procScenario,
                                                       scenarioFunc,
                                                       **scenarioArgs)

        expressTask.setTaskType("Express")

        self.addLogCollectTask(expressTask, taskName="ExpressLogCollect")

        for expressOutLabel, expressOutInfo in viewitems(expressOutMods):

            if expressOutInfo['dataTier'] == "ALCARECO":

                # finalize splitting parameters
                mySplitArgs = self.expressMergeSplitArgs.copy()
                mySplitArgs['algo_package'] = "T0.JobSplitting"

                alcaSkimTask = expressTask.addTask("%sAlcaSkim%s" % (expressTask.name(), expressOutLabel))

                alcaSkimTask.setInputReference(expressTask.getStep(expressRecoStepName),
                                               outputModule=expressOutLabel,
                                               dataTier="ALCARECO")

                alcaTaskConf = {'Multicore': 1,
                                'EventStreams': 0}

                scenarioArgs = {'globalTag': self.globalTag,
                                'skims': self.alcaSkims,
                                'primaryDataset': self.specialDataset}

                if self.globalTagConnect:
                    scenarioArgs['globalTagConnect'] = self.globalTagConnect

                alcaSkimOutMods = self.setupProcessingTask(alcaSkimTask, taskType,
                                                           scenarioName=self.procScenario,
                                                           scenarioFunc="alcaSkim",
                                                           scenarioArgs=scenarioArgs,
                                                           splitAlgo="ExpressMerge",
                                                           splitArgs=mySplitArgs,
                                                           stepType=cmsswStepType,
                                                           forceMerged=True,
                                                           taskConf=alcaTaskConf)

                alcaSkimTask.setTaskType("Express")

                self.addLogCollectTask(alcaSkimTask, taskName="AlcaSkimLogCollect")
                self.addCleanupTask(expressTask, expressOutLabel, dataTier=expressOutInfo['dataTier'])

                for alcaSkimOutLabel, alcaSkimOutInfo in viewitems(alcaSkimOutMods):

                    if alcaSkimOutInfo['dataTier'] == "ALCAPROMPT" and \
                            (self.alcaHarvestCondLFNBase is not None or self.alcaHarvestLumiURL is not None):
                        harvestTask = self.addAlcaHarvestTask(alcaSkimTask, alcaSkimOutLabel,
                                                              alcapromptdataset=alcaSkimOutInfo['filterName'],
                                                              condOutLabel=self.alcaHarvestOutLabel,
                                                              condLFNBase=self.alcaHarvestCondLFNBase,
                                                              lumiURL=self.alcaHarvestLumiURL,
                                                              uploadProxy=self.dqmUploadProxy,
                                                              doLogCollect=True)

                        self.addConditionTask(harvestTask, self.alcaHarvestOutLabel)

            else:

                mergeTask = self.addExpressMergeTask(expressTask, expressRecoStepName, expressOutLabel)

                if expressOutInfo['dataTier'] in ["DQM", "DQMIO"]:
                    self.addDQMHarvestTask(mergeTask, "Merged",
                                           uploadProxy=self.dqmUploadProxy,
                                           periodic_harvest_interval=self.periodicHarvestInterval,
                                           doLogCollect=True,
                                           maxpss=self.harvestingMemory)

        # setting the parameters which need to be set for all the tasks
        # sets acquisitionEra, processingVersion, processingString
        workload.setTaskPropertiesFromWorkload()

        # set the LFN bases (normally done by request manager)
        # also pass run number to add run based directories
        workload.setLFNBase(self.mergedLFNBase, self.unmergedLFNBase,
                            runNumber=self.runNumber)

        return workload

    def addExpressMergeTask(self, parentTask, parentStepName, parentOutputModuleName):
        """
        _addExpressMergeTask_

        Create an expressmerge task for files produced by the parent task

        """
        # finalize splitting parameters
        mySplitArgs = self.expressMergeSplitArgs.copy()
        mySplitArgs['algo_package'] = "T0.JobSplitting"

        parentTaskCmssw = parentTask.getStep(parentStepName)
        parentOutputModule = parentTaskCmssw.getOutputModule(parentOutputModuleName)

        mergeTask = parentTask.addTask("%sMerge%s" % (parentTask.name(), parentOutputModuleName))

        dataTier = getattr(parentOutputModule, "dataTier")
        mergeTask.setInputReference(parentTaskCmssw, outputModule=parentOutputModuleName, dataTier=dataTier)

        self.addRuntimeMonitors(mergeTask)
        mergeTaskCmssw = mergeTask.makeStep("cmsRun1")
        mergeTaskCmssw.setStepType("CMSSW")

        mergeTaskStageOut = mergeTaskCmssw.addStep("stageOut1")
        mergeTaskStageOut.setStepType("StageOut")
        mergeTaskLogArch = mergeTaskCmssw.addStep("logArch1")
        mergeTaskLogArch.setStepType("LogArchive")

        mergeTask.setTaskLogBaseLFN(self.unmergedLFNBase)

        self.addLogCollectTask(mergeTask, taskName="%s%sMergeLogCollect" % (parentTask.name(), parentOutputModuleName))

        mergeTask.applyTemplates()

        mergeTaskCmsswHelper = mergeTaskCmssw.getTypeHelper()

        mergeTaskCmsswHelper.cmsswSetup(self.frameworkVersion, softwareEnvironment="",
                                        scramArch=self.scramArch)

        mergeTaskCmsswHelper.setErrorDestinationStep(stepName=mergeTaskLogArch.name())
        mergeTaskCmsswHelper.setGlobalTag(self.globalTag)
        mergeTaskCmsswHelper.setOverrideCatalog(self.overrideCatalog)

        # mergeTaskStageHelper = mergeTaskStageOut.getTypeHelper()
        # mergeTaskStageHelper.setMinMergeSize(0, 0)

        mergeTask.setTaskType("Merge")

        # DQM is handled differently
        #  merging does not increase size
        #                => disable size limits
        #  only harvest every 15 min
        #                => higher limits for latency (disabled for now)
        dataTier = getattr(parentOutputModule, "dataTier")
        if dataTier in ["DQM", "DQMIO"]:
            mySplitArgs['maxInputSize'] *= 100

        mergeTask.setSplittingAlgorithm("ExpressMerge",
                                        **mySplitArgs)
        mergeTaskCmsswHelper.setDataProcessingConfig(self.procScenario, "merge",
                                                     newDQMIO=(dataTier == "DQMIO"))

        self.addOutputModule(mergeTask, "Merged",
                             primaryDataset=getattr(parentOutputModule, "primaryDataset"),
                             dataTier=getattr(parentOutputModule, "dataTier"),
                             filterName=getattr(parentOutputModule, "filterName"),
                             forceMerged=True)

        self.addCleanupTask(parentTask, parentOutputModuleName, dataTier=getattr(parentOutputModule, "dataTier"))

        return mergeTask

    def addAlcaHarvestTask(self, parentTask, parentOutputModuleName,
                           alcapromptdataset, condOutLabel, condLFNBase, lumiURL,
                           uploadProxy, parentStepName="cmsRun1", doLogCollect=True):
        """
        _addAlcaHarvestTask_

        Create an Alca harvest task to harvest the files produces by the parent task.
        """
        # finalize splitting parameters
        mySplitArgs = {}
        mySplitArgs['algo_package'] = "T0.JobSplitting"
        mySplitArgs['runNumber'] = self.runNumber
        mySplitArgs['alcapromptdataset'] = alcapromptdataset
        mySplitArgs['timeout'] = self.alcaHarvestTimeout

        harvestTask = parentTask.addTask("%sAlcaHarvest%s" % (parentTask.name(), parentOutputModuleName))
        self.addRuntimeMonitors(harvestTask)
        harvestTaskCmssw = harvestTask.makeStep("cmsRun1")
        harvestTaskCmssw.setStepType("CMSSW")

        harvestTaskCondition = harvestTaskCmssw.addStep("condition1")
        harvestTaskCondition.setStepType("AlcaHarvest")
        harvestTaskUpload = harvestTaskCmssw.addStep("upload1")
        harvestTaskUpload.setStepType("DQMUpload")
        harvestTaskLogArch = harvestTaskCmssw.addStep("logArch1")
        harvestTaskLogArch.setStepType("LogArchive")

        harvestTask.setTaskLogBaseLFN(self.unmergedLFNBase)
        if doLogCollect:
            self.addLogCollectTask(harvestTask,
                                   taskName="%s%sAlcaHarvestLogCollect" % (parentTask.name(), parentOutputModuleName))

        harvestTask.setTaskType("Harvesting")
        harvestTask.applyTemplates()

        harvestTaskCmsswHelper = harvestTaskCmssw.getTypeHelper()
        harvestTaskCmsswHelper.cmsswSetup(self.frameworkVersion, softwareEnvironment="",
                                          scramArch=self.scramArch)

        harvestTaskCmsswHelper.setErrorDestinationStep(stepName=harvestTaskLogArch.name())
        harvestTaskCmsswHelper.setGlobalTag(self.globalTag)
        harvestTaskCmsswHelper.setOverrideCatalog(self.overrideCatalog)

        harvestTaskCmsswHelper.setUserLFNBase("/")

        parentTaskCmssw = parentTask.getStep(parentStepName)
        parentOutputModule = parentTaskCmssw.getOutputModule(parentOutputModuleName)

        dataTier = getattr(parentOutputModule, "dataTier")
        harvestTask.setInputReference(parentTaskCmssw, outputModule=parentOutputModuleName, dataTier=dataTier)

        harvestTask.setSplittingAlgorithm("AlcaHarvest",
                                          **mySplitArgs)

        scenarioArgs = {'globalTag': self.globalTag,
                        'datasetName': "/%s/%s/%s" % (getattr(parentOutputModule, "primaryDataset"),
                                                      getattr(parentOutputModule, "processedDataset"),
                                                      dataTier),
                        'runNumber': self.runNumber,
                        'alcapromptdataset': alcapromptdataset}

        if self.globalTagConnect:
            scenarioArgs['globalTagConnect'] = self.globalTagConnect

        harvestTaskCmsswHelper.setDataProcessingConfig(self.procScenario,
                                                       "alcaHarvesting",
                                                       **scenarioArgs)

        harvestTaskConditionHelper = harvestTaskCondition.getTypeHelper()
        harvestTaskConditionHelper.setRunNumber(self.runNumber)
        harvestTaskConditionHelper.setConditionOutputLabel(condOutLabel + "ALCAPROMPT")
        harvestTaskConditionHelper.setConditionLFNBase(condLFNBase)
        harvestTaskConditionHelper.setLuminosityURL(lumiURL)

        self.addOutputModule(harvestTask, condOutLabel,
                             primaryDataset=getattr(parentOutputModule, "primaryDataset"),
                             dataTier=getattr(parentOutputModule, "dataTier"),
                             filterName=getattr(parentOutputModule, "filterName"))

        harvestTaskUploadHelper = harvestTaskUpload.getTypeHelper()
        harvestTaskUploadHelper.setProxyFile(uploadProxy)
        harvestTaskUploadHelper.setServerURL(self.dqmUploadUrl)

        return harvestTask

    def addConditionTask(self, parentTask, parentOutputModuleName):
        """
        _addConditionTask_

        Does not actually produce any jobs
        The job splitter is custom and just forwards information
        into T0AST specific data structures, the actual upload
        of the conditions to the DropBox is handled in a separate
        Tier0 component.

        """
        # finalize splitting parameters
        mySplitArgs = {}
        mySplitArgs['algo_package'] = "T0.JobSplitting"
        mySplitArgs['runNumber'] = self.runNumber
        mySplitArgs['streamName'] = self.streamName

        parentTaskCmssw = parentTask.getStep("cmsRun1")
        parentOutputModule = parentTaskCmssw.getOutputModule(parentOutputModuleName)

        conditionTask = parentTask.addTask("%sCondition%s" % (parentTask.name(), parentOutputModuleName))

        # this is complete bogus, but other code can't deal with a task with no steps
        conditionTaskBogus = conditionTask.makeStep("bogus")
        conditionTaskBogus.setStepType("DQMUpload")

        dataTier = getattr(parentOutputModule, "dataTier")
        conditionTask.setInputReference(parentTaskCmssw, outputModule=parentOutputModuleName, dataTier=dataTier)

        conditionTask.applyTemplates()

        conditionTask.setTaskType("Harvesting")

        conditionTask.setSplittingAlgorithm("Condition",
                                            **mySplitArgs)

        return

    def __call__(self, workloadName, arguments):
        """
        _call_

        Create a Express workload with the given parameters.
        """
        StdBase.__call__(self, workloadName, arguments)

        # Required parameters that must be specified by the Requestor.
        self.outputs = arguments['Outputs']

        # job splitting parameters (also required parameters)
        self.expressSplitArgs = {}
        self.expressSplitArgs['maxInputRate'] = arguments['MaxInputRate']
        self.expressSplitArgs['maxInputEvents'] = arguments['MaxInputEvents']
        self.expressMergeSplitArgs = {}
        self.expressMergeSplitArgs['maxInputSize'] = arguments['MaxInputSize']
        self.expressMergeSplitArgs['maxInputFiles'] = arguments['MaxInputFiles']
        self.expressMergeSplitArgs['maxLatency'] = arguments['MaxLatency']

        # fixed parameters that are used in various places
        self.alcaHarvestOutLabel = "Sqlite"

        return self.buildWorkload()

    @staticmethod
    def getWorkloadCreateArgs():
        baseArgs = StdBase.getWorkloadCreateArgs()
        specArgs = {"RequestType": {"default": "Express"},
                    "ConfigCacheID": {"optional": True, "null": True},
                    "Scenario": {"optional": False, "attr": "procScenario"},
                    "RecoCMSSWVersion": {"validate": lambda x: x in releases(),
                                         "optional": False, "attr": "recoFrameworkVersion"},
                    "RecoScramArch": {"validate": lambda x: all([y in architectures() for y in x]),
                                      "optional": False, "type": makeNonEmptyList},
                    "GlobalTag": {"optional": False},
                    "ProcessingString": {"default": "", "validate": procstringT0},
                    "StreamName": {"optional": False},
                    "SpecialDataset": {"optional": False},
                    "AlcaHarvestTimeout": {"type": int, "optional": False},
                    "AlcaHarvestCondLFNBase": {"optional": False, "null": True},
                    "AlcaHarvestLumiURL": {"optional": False, "null": True},
                    "AlcaSkims": {"type": makeList, "optional": False},
                    "DQMSequences": {"type": makeList, "attr": "dqmSequences", "optional": False},
                    "Outputs": {"type": makeList, "optional": False},
                    "MaxInputRate": {"type": int, "optional": False},
                    "MaxInputEvents": {"type": int, "optional": False},
                    "MaxInputSize": {"type": int, "optional": False},
                    "MaxInputFiles": {"type": int, "optional": False},
                    "MaxLatency": {"type": int, "optional": False},
                    }
        baseArgs.update(specArgs)
        StdBase.setDefaultArgumentsProperty(baseArgs)
        return baseArgs

    @staticmethod
    def getWorkloadAssignArgs():
        baseArgs = StdBase.getWorkloadAssignArgs()
        specArgs = {
            "Override": {"default": {"eos-lfn-prefix": "root://eoscms.cern.ch//eos/cms/store/logs/prod/recent/Express"},
                         "type": dict},
            }
        baseArgs.update(specArgs)
        StdBase.setDefaultArgumentsProperty(baseArgs)
        return baseArgs



#!/usr/bin/env python
"""
_AgentWatchdog_

This component is the default watchdog for WMAgent.

A simple watchdog mechanism based on OS signal handling, implemented to help in the monitoring
of the health of all WMAgent components.
"""

import logging
import threading
import time

from WMCore.Agent.Harness import Harness
from WMComponent.AgentWatchdog.AgentWatchdogPoller import AgentWatchdogPoller
from WMComponent.AgentWatchdog.AgentWatchdogScanner import AgentWatchdogScanner


class AgentWatchdog(Harness):
    """
    Component class for AgentWatchdog module
    """

    def __init__(self, config):
        """
        __init__

        Initialize the Harness
        """
        Harness.__init__(self, config)

        return

    def preInitialization(self):
        """
        _preInitialization_

        Sets up the worker thread
        """

        logging.info("AgentWatchdog.preInitialization")

        # Do not wait between runs.
        # NOTE: For AgentWatchdogPoller, the configured pooling interval will actually be used as a wait
        #       time inside the main thread to drive the cyclic process for regular refresh of all timers.
        #       The result will be, that we will be executing every time for exactly the amount
        #       of time configured as polling cycle in the component config and wait for 0 seconds
        #       between re-execution. In contrast to how all other WMAgent components behave,
        #       where, they execute for a random amount of time (driven by their resource consumption
        #       and other delays during the execution itself), and then pause for the time configured
        #       as component's poll interval before re-executing the main thread algorithm method again.
        #       If we were to follow the order established for the other components, the value here should
        #       have been:
        #       agentWatchdogPollInterval = self.config.AgentWatchdog.pollInterval
        agentWatchdogPollerInterval = 0
        agentWatchdogScannerInterval = self.config.AgentWatchdog.AgentWatchdogScanner.pollInterval
        currThread = threading.currentThread()

        if not hasattr(self.config, "Tier0Feeder"):
            logging.info("Setting AgentWatchdogPoller poll interval to %s seconds", agentWatchdogPollerInterval)
            agentWatchdogPollerThread = currThread.workerThreadManager.addWorker(AgentWatchdogPoller(self.config),
                                                                             agentWatchdogPollerInterval)
            logging.info(f"AgentWatchdog thread PID: {currThread.native_id}")
            logging.info(f"AgentWatchdogPoller thread PID: {agentWatchdogPollerThread.native_id}")

            logging.info("Setting AgentWatchdogScanner poll interval to %s seconds", agentWatchdogScannerInterval)
            agentWatchdogScannerThread = currThread.workerThreadManager.addWorker(AgentWatchdogScanner(self.config),
                                                                             agentWatchdogScannerInterval)
            logging.info(f"AgentWatchdog thread PID: {currThread.native_id}")
            logging.info(f"AgentWatchdogPoller thread PID: {agentWatchdogPollerThread.native_id}")

        return



#!/usr/bin/env python
# pylint: disable=W0613
"""
__RetryManagerPoller__

This component does the actualy retry logic. It allows to have
different algorithms for each job type.

The configuration for this component is as follows:

config.RetryManager.plugins is a dictionary with job types as keys and
plugin names as values, e.g. {'Processing' : 'SquaredAlgo'}. Job types
that don't appear in the keys of the dictionary will use the algorithm
defined for the key 'default' which defaults to DefaultRetryAlgo if
not specified.

For each plugin that appears in the values of config.RetryManager.plugins,
the params can be configured for each job type with the following syntax:

config.RetryManager.<AlgoName>.<JobType>.<ParamName> = <param>

If a particular job type doesn't appear under the corresponding algorithm
configuration, the component will use the params configured for 'default'
or the internal defaults if the default parameters aren't configured either.

An example configuration:

config.RetryManager.plugins = {'Processing' : 'PauseAlgo',
                               'Cleanup' : 'LinearAlgo',
                               'default' : 'SquaredAlgo'}
config.RetryManager.section_('PauseAlgo')
config.RetryManager.PauseAlgo.section_('Processing')
config.RetryManager.PauseAlgo.Processing.coolOffTime = {'submit' : 50, 'create' : 50, 'job' : 20}
config.RetryManager.PauseAlgo.Processing.pauseCount = 3
config.RetryManager.PauseAlgo.section_('default')
config.RetryManager.PauseAlgo.default.coolOffTime = {'submit' : 50, 'create' : 50, 'job' : 20}
config.RetryManager.PauseAlgo.default.pauseCount = 2
config.RetryManager.section_('LinearAlgo')
config.RetryManager.LinearAlgo.section_('Cleanup')
config.RetryManager.LinearAlgo.Cleanup.coolOffTime = {'submit' : 50, 'create' : 50, 'job' : 20}
config.RetryManager.LinearAlgo.section_('default')
config.RetryManager.LinearAlgo.default.coolOffTime = {'submit' : 50, 'create' : 50, 'job' : 20}
config.RetryManager.section_('SquaredAlgo')
config.RetryManager.SquaredAlgo.section_('default')
config.RetryManager.SquaredAlgo.default.coolOffTime = {'submit' : 50, 'create' : 50, 'job' : 20}

Note: It is possible to not specify any configuration at all and the
component won't crash but it won't do anything at all. All
jobs that get in cooloff would stay there forever.
Any parameter can be skipped and the component will use internal defaults.
"""

from future.utils import viewvalues

import datetime
import logging
import threading
import time
import traceback
from Utils.Timers import timeFunction
from Utils.wmcoreDTools import resetWatchdogTimer, moduleName
from WMCore.DAOFactory import DAOFactory
from WMCore.JobStateMachine.ChangeState import ChangeState
from WMCore.JobStateMachine.Transitions import Transitions
from WMCore.WMBS.Job import Job
from WMCore.WMException import WMException
from WMCore.WMFactory import WMFactory
from WMCore.WorkerThreads.BaseWorkerThread import BaseWorkerThread

__all__ = []


def convertdatetime(t):
    """
    Convert dates into useable format.
    """
    return int(time.mktime(t.timetuple()))


def timestamp():
    """
    generate a timestamp
    """
    t = datetime.datetime.now()
    return convertdatetime(t)


class RetryManagerException(WMException):
    """
    _RetryManagerException_

    It's totally awesome, except it's not.
    """


class RetryManagerPoller(BaseWorkerThread):
    """
    _RetryManagerPoller_

    Polls for Jobs in CoolOff State and attempts to retry them
    based on the requirements in the selected plugin
    """

    def __init__(self, config):
        """
        Initialise class members
        """
        BaseWorkerThread.__init__(self)
        self.config = config

        myThread = threading.currentThread()

        self.daoFactory = DAOFactory(package="WMCore.WMBS",
                                     logger=myThread.logger,
                                     dbinterface=myThread.dbi)

        pluginPath = getattr(self.config.RetryManager, "pluginPath",
                             "WMComponent.RetryManager.PlugIns")
        self.pluginFactory = WMFactory("plugins", pluginPath)

        self.changeState = ChangeState(self.config)
        self.getJobs = self.daoFactory(classname="Jobs.GetAllJobs")


        # get needed plugins
        self.plugins = {}

        self.typePluginsAssoc = getattr(self.config.RetryManager, 'plugins', {})
        self.typePluginsAssoc.setdefault('default', 'DefaultRetryAlgo')

        for pluginName in viewvalues(self.typePluginsAssoc):
            try:
                plugin = self.pluginFactory.loadObject(classname=pluginName,
                                                       args=config)
                self.plugins[pluginName] = plugin
            except Exception as ex:
                msg = "Error loading plugin %s on path %s\n" % (pluginName, pluginPath)
                msg += str(ex)
                logging.error(msg)
                raise RetryManagerException(msg)

        return

    def terminate(self, params):
        """
        Run one more time through, then terminate

        """
        logging.debug("Terminating. doing one more pass before we die")
        self.algorithm(params)

    @timeFunction
    def algorithm(self, parameters=None):
        """
        Performs the doRetries method, loading the appropriate
        plugin for each job and handling it.
        """
        logging.debug("Running retryManager algorithm")
        myThread = threading.currentThread()
        try:
            myThread.transaction.begin()
            self.doRetries()
            myThread.transaction.commit()
        except WMException:
            if getattr(myThread, 'transaction', None) and \
                    getattr(myThread.transaction, 'transaction', None):
                myThread.transaction.rollback()
            raise
        except Exception as ex:
            msg = "Caught exception in RetryManager\n"
            msg += str(ex)
            msg += str(traceback.format_exc())
            msg += "\n\n"
            logging.error(msg)
            if hasattr(myThread, 'transaction') \
                    and myThread.transaction is not None \
                    and hasattr(myThread.transaction, 'transaction') \
                    and myThread.transaction.transaction is not None:
                myThread.transaction.rollback()
            raise Exception(msg)

        # Reset its own watchdog timer at the end of the run cycle
        logging.info(f"Resetting {moduleName(self)} watchdog timer.")
        if resetWatchdogTimer(self):
            logging.warning(f"Failed to reset {moduleName(self)} watchdog timer. The component might be restarted soon.")


    def processRetries(self, jobs, cooloffType):
        """
        _processRetries_

        Actually does the dirty work of figuring out what to do with jobs
        """

        if len(jobs) < 1:
            # We got no jobs?
            return

        transitions = Transitions()
        oldstate = '%scooloff' % cooloffType
        if oldstate not in transitions:
            msg = 'Unknown job type %s' % cooloffType
            logging.error(msg)
            return
        propList = []

        newJobState = transitions[oldstate][0]

        jobList = self.loadJobsFromList(idList=jobs)

        # Now we should have the jobs
        propList = self.selectRetryAlgo(jobList, cooloffType)

        if len(propList) > 0:
            self.changeState.propagate(propList, newJobState, oldstate)

        return

    def loadJobsFromList(self, idList):
        """
        _loadJobsFromList_

        Load jobs in bulk
        """

        loadAction = self.daoFactory(classname="Jobs.LoadFromID")
        getTypeAction = self.daoFactory(classname="Jobs.GetType")

        binds = []
        for jobID in idList:
            binds.append({"jobid": jobID})

        results = loadAction.execute(jobID=binds)
        typeResults = getTypeAction.execute(jobID=idList)
        subTypes = {}

        for typeEntry in typeResults:
            subTypes[typeEntry['id']] = typeEntry['type']

        # You have to have a list
        if isinstance(results, dict):
            results = [results]

        listOfJobs = []
        for entry in results:
            # One job per entry
            tmpJob = Job(id=entry['id'])
            tmpJob.update(entry)
            tmpJob['jobType'] = subTypes[entry['id']]
            listOfJobs.append(tmpJob)

        return listOfJobs

    def selectRetryAlgo(self, jobList, cooloffType):
        """
        _selectRetryAlgo_

        Selects which retry algorithm to use
        """
        result = []

        if len(jobList) == 0:
            return result

        for job in jobList:
            try:
                if job['jobType'] in self.typePluginsAssoc:
                    pluginName = self.typePluginsAssoc[job['jobType']]
                else:
                    pluginName = self.typePluginsAssoc['default']
                plugin = self.plugins[pluginName]

                if plugin.isReady(job=job, cooloffType=cooloffType):
                    result.append(job)
            except Exception as ex:
                msg = "Exception while checking for cooloff timeout for job %i\n" % job['id']
                msg += str(ex)
                logging.error(msg)
                logging.debug("Job: %s\n", job)
                logging.debug("cooloffType: %s\n", cooloffType)
                raise RetryManagerException(msg)

        return result

    def doRetries(self):
        """
        Queries DB for all watched filesets, if matching filesets become
        available, create the subscriptions
        """
        # Discover the jobs that are in create cooloff
        jobs = self.getJobs.execute(state='createcooloff')
        logging.info("Found %s jobs in createcooloff", len(jobs))
        self.processRetries(jobs, 'create')

        # Discover the jobs that are in submit cooloff
        jobs = self.getJobs.execute(state='submitcooloff')
        logging.info("Found %s jobs in submitcooloff", len(jobs))
        self.processRetries(jobs, 'submit')

        # Discover the jobs that are in run cooloff
        jobs = self.getJobs.execute(state='jobcooloff')
        logging.info("Found %s jobs in jobcooloff", len(jobs))
        self.processRetries(jobs, 'job')

        # Discover the jobs that are in paused, logging only purpose:
        jobs = self.getJobs.execute(state='jobpaused')
        logging.info("Found %s jobs in jobpaused", len(jobs))

        jobs = self.getJobs.execute(state='createpaused')
        logging.info("Found %s jobs in createpaused", len(jobs))

        jobs = self.getJobs.execute(state='submitpaused')
        logging.info("Found %s jobs in submitpaused", len(jobs))



#!/usr/bin/env python
"""
_StageOut.Impl_

Implementations of the StageOutImpl for specific protocols

"""


__all__ = []

#  //
# // Each plugin should contain the Registration call at module level
#//  and be imported here, so that all the StageOut plugins can be
#  //imported automatically
# //
#//



"""
_MarkBlocksDeleted_

MySQL implementation of PhEDExInjector.MarkBlocksDeleted

Set deleted status for blocks

"""

from __future__ import division
from __future__ import print_function

from WMCore.Database.DBFormatter import DBFormatter


class MarkBlocksDeleted(DBFormatter):
    sql = """UPDATE dbsbuffer_block
             SET dbsbuffer_block.deleted = :DELETED
             WHERE dbsbuffer_block.blockname = :BLOCKNAME
             """

    def execute(self, binds, conn=None, transaction=False):

        if not binds:
            return

        self.dbi.processData(self.sql, binds, conn=conn,
                             transaction=transaction)

        return



#!/usr/bin/env python
"""
_RucioInjector_

Poll the DBSBuffer tables and insert files, blocks and datasets into Rucio
"""
from __future__ import division

import logging
import threading

from WMComponent.RucioInjector.RucioInjectorPoller import RucioInjectorPoller
from WMCore.Agent.Harness import Harness


class RucioInjector(Harness):
    def __init__(self, config):
        Harness.__init__(self, config)

    def preInitialization(self):
        pollInterval = self.config.RucioInjector.RucioInjectorPoller.pollInterval
        logging.info("Setting poll interval to %s seconds for inject", pollInterval)

        myThread = threading.currentThread()
        myThread.workerThreadManager.addWorker(RucioInjectorPoller(self.config), pollInterval)



"""
_GetPSNtoPNNMapping_

Oracle implementation of Locations.GetPSNtoPNNMapping

"""
from __future__ import (print_function, division)

from WMCore.WMBS.MySQL.Locations.GetPSNtoPNNMapping import GetPSNtoPNNMapping as MySQLGetPSNtoPNNMapping

class GetPSNtoPNNMapping(MySQLGetPSNtoPNNMapping):
    """
    Same as MySQL version

    """
    pass



#!/usr/bin/env python
"""
Retrieve basic information about the worker threads such
that it can be pushed to the monitoring tool
"""
from __future__ import division

from WMCore.Agent.Database.MySQL.MonitorWorkers import MonitorWorkers \
    as MySQLMonitorWorkers


class MonitorWorkers(MySQLMonitorWorkers):
    pass



"""
A simple watchdog mechanism based on OS signal handling, implemented to help in the monitoring
of the health of all WMAgent components.

This AgentWatchdogScanner thread is to be scanning all other components' process trees
and will take the proper (restart) actions in case of unhealthy component is found.
While at the same time the AgentWatchdogPoller is to be endowed with a mechanism
to monitor for run time overflow in all agent components and will take care of
proper restarts of the misbehaved ones.


And we shall call this component with the affectionate nickname: Cerberus ;)
"""

import threading
import os
import signal
import logging
import time
import subprocess
import psutil
import inspect
import random
import json
import re

from pprint import pformat

from Utils.ProcFS import processStatus
from Utils.Timers import timeFunction
from Utils.wmcoreDTools import getComponentThreads, restart, forkRestart, isComponentAlive
from WMComponent.AgentWatchdog.Timer import Timer, _countdown, WatchdogAction, TimerException
from WMCore.WorkerThreads.BaseWorkerThread import BaseWorkerThread
from WMCore.WMInit import connectToDB
from WMCore.Services.AlertManager.AlertManagerAPI import AlertManagerAPI


class AgentWatchdogScanner(BaseWorkerThread):
    """
    The basic AgentWatchdogScanner class
    """
    def  __init__(self, config):
        BaseWorkerThread.__init__(self)
        self.config = config
        self.mainThread = threading.currentThread()

        # self.watchedComponents = self.config.AgentWatchdog.watchedComponents
        # self.actionLimit = self.config.AgentWatchdog.actionLimit
        self.pollInterval = self.config.AgentWatchdog.AgentWatchdogScanner.pollInterval

        self.mainThread.parents = [ thread['pid'] for thread in processStatus(self.mainThread.native_id)]
        logging.info(f"Initialized with parents {self.mainThread.parents}.")
        self.mainThread.enum = [ pid.native_id for pid in threading.enumerate() ]
        logging.info(f"Initialized with the following threads: {self.mainThread.enum}.")
        self.mainThread.main = threading.main_thread().native_id
        logging.debug(f"Initialized with main_thread {self.mainThread.main}.")

        self.alertManagerUrl = self.config.Alert.alertManagerUrl
        self.alertDestinationMap = self.config.Alert.alertDestinationMap
        self.alertManager = AlertManagerAPI(self.alertManagerUrl)
        logging.info(f"Setting up an alertManager instance for AgentWatchdogScanner and redirecting alerts to: {self.alertManagerUrl}.")

    def sendAlert(self, alertMessage, severity='low'):
        """
        A generic method for sending alerts from AgentWatchdogScanner
        :param alertMesasge: A string with the alert contents
        :param severity:     Default: 'low'
        NOTE: The rest of the alertAPI parameters come from the agent configuration at init time
        """
        currAgent = getattr(self.config.Agent, 'hostName')
        alertName = "AgentWatchdogScanner"
        summary = f"Alert from WMAgent {currAgent}"
        description = alertMessage
        service = f"AgentWatchdogScanner@{currAgent}"
        self.alertManager.sendAlert(alertName, severity, summary, description, service, tag=self.alertDestinationMap['alertAgentWatchdogScanner'])

    def checkCompAlive(self):
        """
        # Iterate through all components and check if they have a healthy process tree:
        # Restart any of them found to have problems.
        # NOTE: Exclude AgentWatchdog and itself from this logic, because:
        #       * AgentWatchdog should never be touched since it monitors all others poling cycles
        #       * AgentWatchdog is going to spawn threads dynamically and will fail all checks
        #       * AgentWatchdog cannot restart itself
        """
        logging.info(f"Checking all components' threads.")
        for component in [comp for comp in self.config.listComponents_() if comp != 'AgentWatchdog']:
            if not isComponentAlive(self.config, component=component):
                try:
                    forkRestart(self.config, componentsList=[component])
                    msg = f"Restarted Unhealthy component: {component}"
                    logging.warning(f"{msg}")
                    self.sendAlert(msg)
                except Exception as ex:
                    logging.error(f"Failed to restart component: {component}. Full ERROR: {str(ex)}")
                    raise

    def setup(self, parameters):
        # Wait one full cycle before starting the whole component to allow all others to initialize properly
        logging.info("Waiting for 10s to let others initialize properly ... ")

        time.sleep(10)
        logging.info("Start Monitoring the components' threads ... ")

    @timeFunction
    def algorithm(self, parameters=None):
        """
        Spawn and register watchdog timers
        """
        # TODO: To implement logic with limited number of restarts, similar to AgentWatchdogPoller
        try:
            currThread = threading.currentThread()
            logging.info(f"Polling cycle started with current pid: {currThread.native_id}.")
            self.checkCompAlive()
        except Exception as ex:
            logging.exception(f"Exception occurred while executing the AgentWatchdogScanner logic.")



import signal
import logging
import time
import inspect
import json
import threading

from collections  import namedtuple
from typing import NamedTuple
from collections.abc import Callable
from threading import Thread
from pprint import pformat


def _countdown(endTime):
    """
    _countdown_

    Aux function to return the remaining time before reaching the endTime
    :param endTime: The end time in seconds since the epoch.
    :return:        Remaining time in seconds or 0 if the endTime has already passed
                    No negative values are returned
    """
    remTime = endTime - time.time()
    if remTime < 0:
        remTime = 0
    return remTime

# NOTE: When an instance of the WatchdogAction tuple is to be created, it must be of the form:
#       action = WatchdogAction(callbackFunction, [args], {kwArgs})

WatchdogAction = NamedTuple('WatchdogAction', [('func', Callable[...,None]),
                                               ('args',list),
                                               ('kwArgs', dict)])


class TimerException(Exception):
    """
    __TimerException__
    Exception raised by the agent related to AgentWatchdog timers.
    """
    def __init__(self, message=None, errorNo=None, **data):
        self.name = str(self.__class__.__name__)
        Exception.__init__(self, self.name, message)


class Timer(Thread):
    """
    Watchdog timer class.
    All instances of this class are about to spawn a separate thread and are to
    expect a list of particular PIDs to reset the timer. If the timer reaches to the end
    of its interval, without being reset, it would call the action function provided
    at initialization time with the respective arguments.
    """
    def __init__(self,
                 name=None,
                 compName=None,
                 action=None,
                 expPids=[],
                 expSig=None,
                 path=None,
                 interval=0,
                 actionLimit=1):
        """
        __init__
        :param expPids:     The list of expected pids allowed to reset the timer, signals from anybody else would be ignored
        :param compName:    The component name this timer is associated with
        :param expSig:      The default signal to be expected for resetting the timer (Default: SIGCONT)
        :param interval:    The interval for the timer
        :param path:        The path where the timer's data is to be preserved on disk
        :param action:      The action to be taken in case a timer gets expired. An instance of WatchdogAction
        :param actionLimit: The maximum number of times an action would be executed
                            (hence, the maximum number of times the timer is allowed to expire) before the timer's thread gets destroyed.
                            Default: 1 - Meaning the timer would be allowed to expire exactly once.
                            NOTE: The action will be executed  at least once, the moment when the timer gets expired for the first time.
        """
        Thread.__init__(self)
        self.name = name
        self.compName = compName
        self.interval = interval
        self.endTime = 0
        self.startTime = 0
        self.expPids = expPids
        self.expSig = expSig or signal.SIGCONT
        self.path = path
        self.daemon = True
        self.action = action
        self.actionString = action.__str__()
        self.actionCounter = 0
        self.actionLimit = actionLimit
        if self.action:
            # First, make sure the signature of the action defined for this timer matches the arguments provided:
            try:
                actionSignature = inspect.signature(self.action.func)
                actionSignature.bind(*self.action.args, **self.action.kwArgs)
            except TypeError as ex:
                msg = f"{self.name}:  The timer's action method signature does not match the set of arguments provided. Error: {str(ex)}"
                logging.exception(msg)
                raise TimerException(msg) from None
        else:
            logging.warning(f"{self.name}: This is a timer with no action defined. This timer should be used mostly for debugging purposes.")

        # This MUST be overwritten at runtime by the thread/process creating the timer
        self.creator_id = None

    def run(self):
        """
        _run_
        Thread class run() method override
        """
        self._timer()

    def write(self):
        """
        _write_
        Method to write the current timer's state on disk:
        :return: Nothing. Logs an error in case the timer was not written correctly
                 No exceptions would be raised
        """
        try:
            with open(self.path, 'w') as timerFile:
                json.dump(self.dictionary_(), timerFile , indent=4)
        except Exception as ex:
            logging.error(f"{self.name}: Failed to write timer data on disk. Timer path: {self.path}. ERROR: {str(ex)}")

    # NOTE: The bellow code for fetching the parent_id may produce wrong results for WMAgent components,
    #       because we do os.fork() when creating our component daemons. This means that if self.creator_id
    #       has not been set at runtime by the thread/process which created this timer the method here
    #       will try to return the native_id of the process which has started the python interpreter,
    #       but at the moment when main_thread() is called, this process is long gone.
    @property
    def parent_id(self):
        return self.creator_id or threading.main_thread().native_id

    @property
    def remTime(self):
        """
        _remTime_
        Returns the current timer's remaining time (in seconds) before the timer's action is to be applied
        """
        return _countdown(self.endTime)

    @property
    def alive(self):
        """
        Returns the current thread status as a property, so that it can be serialized and written to the timer file
        """
        return self.is_alive()

    @staticmethod
    def _isSerializable(obj):
        """
        __isSerializable__
        Auxiliary function to check for object serialization
        :param obj: Object of any type to be checked
        :return:    Bool - True if the object is serializable, False otherwise
        """
        try:
            json.dumps(obj)
            # logging.debug(f"{obj} is serializable")
            return True
        except TypeError:
            # logging.debug(f"{obj} is NOT serializable")
            return False

    def dictionary_(self):
        """
        _dictionary__
        Returns a dictionary representation of the current timer
        """
        return {attr: value for attr,value in inspect.getmembers(self) if self._isSerializable(value)}

    def reset(self, timer=True, actionCounter=False):
        """
        _reset_
        Resets the timer by delaying its endTime with one timer's interval and zeroing the actionCounter
        :param timer:         Bool flag to trigger time reset (Default: True)
        :param actionCounter: Bool flag to trigger actionCounter reset (Default: False)
        """
        try:
            if timer:
                self.endTime = time.time() + self.interval
            if actionCounter:
                self.actionCounter = 0
        except Exception as ex:
            logging.exception(f"{self.name}: pid: {self.native_id}: ERROR while resetting timer: {self.name}!")

    def restart(self, *args, **kwArgs):
        """
        _restart_
        A method allowing a complete reconfiguration and restart of the current timer.
        :param *: Accepts all keyword parameters allowed at __init__

        NOTE: This method merges any newly provided kwArgs with the already
              existing object parameters. Any non kwArgs are ignored
        """
        try:
            # First, check if the timer is still alive and did not yet reach its end of time
            if self.is_alive():
                logging.warning(f"{self.name}: pid: {self.native_id}: You cannot restart a running timer. You should wait until all its action retries get exhausted and the timer's thread is stopped normally.")
                return

            # Second, reconfigure it with the new  set of parameters  and recreate its thread
            logging.info(f"{self.name}: pid: {self.native_id}: Re-configuring timer: {self.name}")
            for arg in inspect.signature(self.__init__).parameters:
                if arg not in kwArgs:
                    kwArgs[arg] = getattr(self, arg, None)
            self.__init__(**kwArgs)

            # Finally, rerun it
            logging.info(f"{self.name}: pid: {self.native_id}: Restarting timer: {self.name}")
            self.start()
            self.write()
        except Exception as ex:
            logging.exception(f"{self.name}: pid: {self.native_id}: ERROR while restarting timer: {self.name}!")

    def update(self, *args, **kwArgs):
        """
        _update_
        A method allowing a complete reconfiguration of the current timer.
        :param *: Accepts all keyword parameters allowed at __init__

        NOTE: This method merges any newly provided kwArgs with the already
              existing object parameters. Any non kwArgs are ignored
        """
        try:
            logging.info(f"{self.name}: pid: {self.native_id}: Updating timer: {self.name} ")
            for arg in kwArgs:
                setattr(self, arg, kwArgs[arg])
            self.write()
        except Exception as ex:
            logging.exception(f"{self.name}: pid: {self.native_id}: ERROR while writing timer: {self.name}!")

    def _timer(self):
        """
        A simple timer method, which will be used to override the main Thread class run() method.
        It would expect one particular PID to reset itself. If the timer reaches to the end of
        its interval, without being reset, it would restart the component it is associated with.
        :return:         Nothing
        """

        self.startTime = time.time()
        self.endTime = self.startTime + self.interval

        while True:
            sigInfo = signal.sigtimedwait([self.expSig], self.remTime)
            if sigInfo:
                logging.info(f"{self.name}: pid: {self.native_id}: Received signal: {pformat(sigInfo)}")
                if sigInfo.si_pid in self.expPids:
                    # Resetting the timer starting again from the current time
                    logging.info(f"{self.name}: pid: {self.native_id}: Resetting timer")
                    # reset both time and actionCounter during this reset:
                    self.reset(timer=True, actionCounter=True)
                    self.write()
                else:
                    # Continue to wait for signal from the correct origin
                    logging.info(f"{self.name}: pid: {self.native_id}: Continue to wait for signal from the correct origin. Remaining time: {self.remTime}")
                    continue
            else:
                logging.info(f"{self.name}: pid: {self.native_id}: Reached the end of timer. Applying action: {self.action}")
                try:
                    self.action.func(*self.action.args, **self.action.kwArgs)
                    # preserve current actionCounter during this reset:
                    self.reset(timer=True, actionCounter=False)
                    self.actionCounter += 1
                    self.write()
                    logging.info(f"{self.name}: pid: {self.native_id}: Current actionCounter: {self.actionCounter}")
                except Exception as ex:
                    currFrame = inspect.currentframe()
                    argsInfo = inspect.getargvalues(currFrame)
                    argVals = {arg: argsInfo.locals.get(arg) for arg in argsInfo.args}
                    msg = f"Failure while applying {self.name} timer's action: {self.action.func.__name__} "
                    msg += f"With arguments: {argVals}"
                    msg += f"Full exception string: {str(ex)}"
                    logging.exception(msg)
                    raise TimerException(msg) from None
                if self.actionCounter >= self.actionLimit:
                    break

        logging.info(f"{self.name}: pid: {self.native_id}: actionLimit of: {self.actionLimit} action repetitions exhausted!")
        logging.info(f"{self.name}: pid: {self.native_id}: Reached the end of timer logic. The timer thread will end now. You can restart it through: timer.restart() !!!")



"""
_NewSubscription_

Oracle implementation of DBS3Buffer.NewSubscription

Created on May 2, 2013

@author: dballest
"""

from WMComponent.DBS3Buffer.MySQL.NewSubscription import NewSubscription as MySQLNewSubscription

class NewSubscription(MySQLNewSubscription):
    """
    _NewSubscription_

    Create a new subscription in the database
    """

    sql = """INSERT INTO dbsbuffer_dataset_subscription
             (dataset_id, site, custodial, priority, subscribed, delete_blocks, dataset_lifetime)
             SELECT :id, :site, :custodial,
                    :priority, 0, :delete_blocks, :dataset_lifetime
             FROM DUAL
             WHERE NOT EXISTS
               ( SELECT *
                 FROM dbsbuffer_dataset_subscription
                 WHERE dataset_id = :id
                 AND site = :site
                 AND custodial = :custodial
                 AND priority = :priority
                 AND dataset_lifetime = :dataset_lifetime )
             """



#!/usr/bin/env python
# pylint: disable=E1103, E1101, C0301
# E1103: Use DB objects attached to thread
# E1101: Create config sections
# C0301: The names for everything are so ridiculously long
# that I'm disabling this.  The rest of you will have to get
# bigger monitors.
"""
_AccountantWorker_

Used by the JobAccountant to do the actual processing of completed jobs.
"""

import collections
import gc
import logging
import os
import threading

from WMComponent.DBS3Buffer.DBSBufferFile import DBSBufferFile
from WMCore.ACDC.DataCollectionService import DataCollectionService
from WMCore.DAOFactory import DAOFactory
from WMCore.Database.CMSCouch import CouchServer
from WMCore.FwkJobReport.Report import Report
from WMCore.JobStateMachine.ChangeState import ChangeState
from WMCore.Services.WMStats.WMStatsWriter import WMStatsWriter
from WMCore.WMBS.File import File
from WMCore.WMBS.Job import Job
from WMCore.WMConnectionBase import WMConnectionBase
from WMCore.WMException import WMException


class AccountantWorkerException(WMException):
    """
    _AccountantWorkerException_

    WMException based specific class
    """


class AccountantWorker(WMConnectionBase):
    """
    Class that actually does the work of parsing FWJRs for the Accountant
    Run through ProcessPool
    """

    def __init__(self, config):
        """
        __init__

        Create all DAO objects that are used by this class.
        """
        WMConnectionBase.__init__(self, "WMCore.WMBS")
        myThread = threading.currentThread()
        self.dbsDaoFactory = DAOFactory(package="WMComponent.DBS3Buffer",
                                        logger=myThread.logger,
                                        dbinterface=myThread.dbi)

        self.getOutputMapAction = self.daofactory(classname="Jobs.GetOutputMap")
        self.bulkAddToFilesetAction = self.daofactory(classname="Fileset.BulkAddByLFN")
        self.bulkParentageAction = self.daofactory(classname="Files.AddBulkParentage")
        self.getJobTypeAction = self.daofactory(classname="Jobs.GetType")
        self.getParentInfoAction = self.daofactory(classname="Files.GetParentAndGrandParentInfo")
        self.setParentageByJob = self.daofactory(classname="Files.SetParentageByJob")
        self.setParentageByMergeJob = self.daofactory(classname="Files.SetParentageByMergeJob")
        self.setFileRunLumi = self.daofactory(classname="Files.AddRunLumi")
        self.setFileLocation = self.daofactory(classname="Files.SetLocationByLFN")
        self.setFileAddChecksum = self.daofactory(classname="Files.AddChecksumByLFN")
        self.addFileAction = self.daofactory(classname="Files.Add")
        self.jobCompleteInput = self.daofactory(classname="Jobs.CompleteInput")
        self.setBulkOutcome = self.daofactory(classname="Jobs.SetOutcomeBulk")
        self.getWorkflowSpec = self.daofactory(classname="Workflow.GetSpecAndNameFromTask")
        self.getJobInfoByID = self.daofactory(classname="Jobs.LoadFromID")
        self.getFullJobInfo = self.daofactory(classname="Jobs.LoadForErrorHandler")
        self.getJobTaskNameAction = self.daofactory(classname="Jobs.GetFWJRTaskName")
        self.pnn2Psn = self.daofactory(classname="Locations.GetPNNtoPSNMapping").execute()

        self.dbsStatusAction = self.dbsDaoFactory(classname="DBSBufferFiles.SetStatus")
        self.dbsParentStatusAction = self.dbsDaoFactory(classname="DBSBufferFiles.GetParentStatus")
        self.dbsChildrenAction = self.dbsDaoFactory(classname="DBSBufferFiles.GetChildren")
        self.dbsCreateFiles = self.dbsDaoFactory(classname="DBSBufferFiles.Add")
        self.dbsSetLocation = self.dbsDaoFactory(classname="DBSBufferFiles.SetLocationByLFN")
        self.dbsInsertLocation = self.dbsDaoFactory(classname="DBSBufferFiles.AddLocation")
        self.dbsSetChecksum = self.dbsDaoFactory(classname="DBSBufferFiles.AddChecksumByLFN")
        self.dbsSetRunLumi = self.dbsDaoFactory(classname="DBSBufferFiles.AddRunLumi")
        self.dbsGetWorkflow = self.dbsDaoFactory(classname="ListWorkflow")

        self.dbsLFNHeritage = self.dbsDaoFactory(classname="DBSBufferFiles.BulkHeritageParent")

        self.stateChanger = ChangeState(config)

        # Decide whether or not to attach jobReport to returned value
        self.returnJobReport = getattr(config.JobAccountant, 'returnReportFromWorker', False)

        # Store location for the specs for DBS
        self.specDir = getattr(config.JobAccountant, 'specDir', None)

        # maximum RAW EDM size for Repack output before data is put into Error dataset and skips PromptReco
        self.maxAllowedRepackOutputSize = getattr(config.JobAccountant, 'maxAllowedRepackOutputSize',
                                                  12 * 1024 * 1024 * 1024)

        # ACDC service
        self.dataCollection = DataCollectionService(url=config.ACDC.couchurl,
                                                    database=config.ACDC.database)

        self.jobDBName = config.JobStateMachine.couchDBName
        self.jobCouchdb = CouchServer(config.JobStateMachine.couchurl)
        self.fwjrCouchDB = None
        self.localWMStats = WMStatsWriter(config.TaskArchiver.localWMStatsURL, appName="WMStatsAgent")

        # Hold data for later commital
        self.dbsFilesToCreate = []
        self.wmbsFilesToBuild = []
        self.wmbsMergeFilesToBuild = []
        self.mergedOutputFiles = []
        self.listOfJobsToSave = []
        self.listOfJobsToFail = []
        self.filesetAssoc = []
        self.parentageBinds = []
        self.parentageBindsForMerge = []
        self.jobsWithSkippedFiles = {}
        self.count = 0
        self.datasetAlgoID = collections.deque(maxlen=1000)
        self.datasetAlgoPaths = collections.deque(maxlen=1000)
        self.dbsLocations = set()
        self.workflowIDs = collections.deque(maxlen=1000)
        self.workflowPaths = collections.deque(maxlen=1000)

        return

    def reset(self):
        """
        _reset_

        Reset all global vars between runs.
        """
        self.dbsFilesToCreate = []
        self.wmbsFilesToBuild = []
        self.wmbsMergeFilesToBuild = []
        self.mergedOutputFiles = []
        self.listOfJobsToSave = []
        self.listOfJobsToFail = []
        self.filesetAssoc = []
        self.parentageBinds = []
        self.parentageBindsForMerge = []
        self.jobsWithSkippedFiles = {}
        gc.collect()
        return

    def loadJobReport(self, jobReportPath):
        """
        _loadJobReport_

        Given a framework job report on disk, load it and return a
        FwkJobReport instance.  If there is any problem loading or parsing the
        framework job report return None.
        """
        # The jobReportPath may be prefixed with "file://" which needs to be
        # removed so it doesn't confuse the FwkJobReport() parser.
        if not jobReportPath:
            logging.error("Bad FwkJobReport Path: %s", jobReportPath)
            return self.createMissingFWKJR(99999, "FWJR path is empty")

        jobReportPath = jobReportPath.replace("file://", "")
        if not os.path.exists(jobReportPath):
            logging.error("Bad FwkJobReport Path: %s", jobReportPath)
            return self.createMissingFWKJR(99999, 'Cannot find file in jobReport path: %s' % jobReportPath)

        if os.path.getsize(jobReportPath) == 0:
            logging.error("Empty FwkJobReport: %s", jobReportPath)
            return self.createMissingFWKJR(99998, 'jobReport of size 0: %s ' % jobReportPath)

        jobReport = Report()

        try:
            jobReport.load(jobReportPath)
        except UnicodeDecodeError:
            logging.error("Hit UnicodeDecodeError exception while loading jobReport: %s", jobReportPath)
            return self.createMissingFWKJR(99997, 'Found undecodable data in jobReport: {}'.format(jobReportPath))
        except Exception as ex:
            msg = "Error loading jobReport: {}\nDetails: {}".format(jobReportPath, str(ex))
            logging.error(msg)
            return self.createMissingFWKJR(99997, 'Cannot load jobReport')

        if not jobReport.listSteps():
            logging.error("FwkJobReport with no steps: %s", jobReportPath)
            return self.createMissingFWKJR(99997, 'jobReport with no steps: %s ' % jobReportPath)

        return jobReport

    def isTaskExistInFWJR(self, jobReport, jobStatus):
        """
        If taskName is not available in the FWJR, then tries to
        recover it getting data from the SQL database.
        """
        if not jobReport.getTaskName():
            logging.warning("Trying to recover a corrupted FWJR for a %s job with job id %s", jobStatus,
                            jobReport.getJobID())
            jobInfo = self.getJobTaskNameAction.execute(jobId=jobReport.getJobID(),
                                                        conn=self.getDBConn(),
                                                        transaction=self.existingTransaction())

            jobReport.setTaskName(jobInfo['taskName'])
            jobReport.save(jobInfo['fwjr_path'])
            if not jobReport.getTaskName():
                msg = "Report to developers. Failed to recover corrupted fwjr for %s job id %s" % (jobStatus,
                                                                                                   jobReport.getJobID())
                raise AccountantWorkerException(msg)
            else:
                logging.info("TaskName '%s' successfully recovered and added to fwjr id %s.", jobReport.getTaskName(),
                             jobReport.getJobID())

        return

    def __call__(self, parameters):
        """
        __call__

        Handle a completed job.  The parameters dictionary will contain the job
        ID and the path to the framework job report.
        """
        returnList = []
        self.reset()

        for job in parameters:
            logging.info("Handling %s", job["fwjr_path"])

            # Load the job and set the ID
            fwkJobReport = self.loadJobReport(job["fwjr_path"])
            fwkJobReport.setJobID(job['id'])

            jobSuccess = self.handleJob(jobID=job["id"],
                                        fwkJobReport=fwkJobReport)

            if self.returnJobReport:
                returnList.append({'id': job["id"], 'jobSuccess': jobSuccess,
                                   'jobReport': fwkJobReport})
            else:
                returnList.append({'id': job["id"], 'jobSuccess': jobSuccess})

            self.count += 1

        existingTransaction = self.beginTransaction()

        # Now things done at the end of the job
        # Do what we can with WMBS files
        self.handleWMBSFiles(self.wmbsFilesToBuild, self.parentageBinds)

        # handle merge files separately since parentage need to set
        # separately to support robust merge
        self.handleWMBSFiles(self.wmbsMergeFilesToBuild, self.parentageBindsForMerge)

        # Create DBSBufferFiles
        self.createFilesInDBSBuffer()

        # Handle filesetAssoc
        if self.filesetAssoc:
            self.bulkAddToFilesetAction.execute(binds=self.filesetAssoc,
                                                conn=self.getDBConn(),
                                                transaction=self.existingTransaction())

        # Move successful jobs to successful
        if self.listOfJobsToSave:
            idList = [x['id'] for x in self.listOfJobsToSave]
            outcomeBinds = [{'jobid': x['id'], 'outcome': x['outcome']} for x in self.listOfJobsToSave]
            self.setBulkOutcome.execute(binds=outcomeBinds,
                                        conn=self.getDBConn(),
                                        transaction=self.existingTransaction())

            self.jobCompleteInput.execute(id=idList,
                                          lfnsToSkip=self.jobsWithSkippedFiles,
                                          conn=self.getDBConn(),
                                          transaction=self.existingTransaction())
            self.stateChanger.propagate(self.listOfJobsToSave, "success", "complete")

        # If we have failed jobs, fail them
        if self.listOfJobsToFail:
            outcomeBinds = [{'jobid': x['id'], 'outcome': x['outcome']} for x in self.listOfJobsToFail]
            self.setBulkOutcome.execute(binds=outcomeBinds,
                                        conn=self.getDBConn(),
                                        transaction=self.existingTransaction())
            self.stateChanger.propagate(self.listOfJobsToFail, "jobfailed", "complete")

        # Arrange WMBS parentage
        if self.parentageBinds:
            self.setParentageByJob.execute(binds=self.parentageBinds,
                                           conn=self.getDBConn(),
                                           transaction=self.existingTransaction())
        if self.parentageBindsForMerge:
            self.setParentageByMergeJob.execute(binds=self.parentageBindsForMerge,
                                                conn=self.getDBConn(),
                                                transaction=self.existingTransaction())

        # Straighten out DBS Parentage
        if self.mergedOutputFiles:
            self.handleDBSBufferParentage()

        if self.jobsWithSkippedFiles:
            self.handleSkippedFiles()

        self.commitTransaction(existingTransaction)

        return returnList

    def outputFilesetsForJob(self, outputMap, merged, moduleLabel, datatier):
        """
        _outputFilesetsForJob_

        Determine if the file should be placed in any other fileset.  Note that
        this will not return the JobGroup output fileset as all jobs will have
        their output placed there.
        """
        # output map identifier uses output module + datatier
        moduleLabel += datatier
        if moduleLabel not in outputMap:
            logging.info("Output module label missing from output map.")
            return []

        outputFilesets = []
        for outputFileset in outputMap[moduleLabel]:
            if merged is False and outputFileset["output_fileset"] is not None:
                outputFilesets.append(outputFileset["output_fileset"])
            else:
                if outputFileset["merged_output_fileset"] is not None:
                    outputFilesets.append(outputFileset["merged_output_fileset"])

        return outputFilesets

    def addFileToDBS(self, jobReportFile, task, errorDataset=False):
        """
        _addFileToDBS_

        Add a file that was output from a job to the DBS buffer.
        """
        datasetInfo = jobReportFile["dataset"]

        dbsFile = DBSBufferFile(lfn=jobReportFile["lfn"],
                                size=jobReportFile["size"],
                                events=jobReportFile["events"],
                                checksums=jobReportFile["checksums"],
                                status="NOTUPLOADED",
                                inPhedex=0)
        dbsFile.setAlgorithm(appName=datasetInfo["applicationName"],
                             appVer=datasetInfo["applicationVersion"],
                             appFam=jobReportFile["module_label"],
                             psetHash="GIBBERISH",
                             configContent=jobReportFile.get('configURL'))

        if errorDataset:
            dbsFile.setDatasetPath("/%s/%s/%s" % (datasetInfo["primaryDataset"] + "-Error",
                                                  datasetInfo["processedDataset"],
                                                  datasetInfo["dataTier"]))
        else:
            dbsFile.setDatasetPath("/%s/%s/%s" % (datasetInfo["primaryDataset"],
                                                  datasetInfo["processedDataset"],
                                                  datasetInfo["dataTier"]))

        dbsFile.setValidStatus(validStatus=jobReportFile.get("validStatus", None))
        dbsFile.setProcessingVer(ver=jobReportFile.get('processingVer', 0))
        dbsFile.setAcquisitionEra(era=jobReportFile.get('acquisitionEra', None))
        dbsFile.setGlobalTag(globalTag=jobReportFile.get('globalTag', None))
        # TODO need to find where to get the prep id
        dbsFile.setPrepID(prep_id=jobReportFile.get('prep_id', None))
        dbsFile['task'] = task
        dbsFile['runs'] = jobReportFile['runs']

        dbsFile.setLocation(pnn=list(jobReportFile["locations"])[0], immediateSave=False)
        self.dbsFilesToCreate.append(dbsFile)
        return

    def findDBSParents(self, lfn):
        """
        _findDBSParents_

        Find the parent of the file in DBS
        This is meant to be called recursively
        """
        parentsInfo = self.getParentInfoAction.execute([lfn],
                                                       conn=self.getDBConn(),
                                                       transaction=self.existingTransaction())
        newParents = set()
        logging.debug("Found %d potential parents for lfn: %s", len(parentsInfo), lfn)
        for parentInfo in parentsInfo:
            # This will catch straight to merge files that do not have redneck
            # parents.  We will mark the straight to merge file from the job
            # as a child of the merged parent.
            if int(parentInfo["merged"]) == 1 and not parentInfo["lfn"].startswith("/store/unmerged/"):
                newParents.add(parentInfo["lfn"])

            elif parentInfo['gpmerged'] is None:
                continue

            # Handle the files that result from merge jobs that aren't redneck
            # children.  We have to setup parentage and then check on whether or
            # not this file has any redneck children and update their parentage
            # information.
            elif int(parentInfo["gpmerged"]) == 1 and not parentInfo["gplfn"].startswith("/store/unmerged/"):
                newParents.add(parentInfo["gplfn"])

            # If that didn't work, we've reached the great-grandparents
            # And we have to work via recursion
            else:
                parentSet = self.findDBSParents(lfn=parentInfo['gplfn'])
                for parent in parentSet:
                    newParents.add(parent)

        return newParents

    def addFileToWMBS(self, jobType, fwjrFile, jobMask, task, jobID=None):
        """
        _addFileToWMBS_

        Add a file that was produced in a job to WMBS.
        """
        fwjrFile["first_event"] = jobMask["FirstEvent"]

        if fwjrFile["first_event"] is None:
            fwjrFile["first_event"] = 0

        if jobType == "Merge" and fwjrFile["module_label"] != "logArchive":
            setattr(fwjrFile["fileRef"], 'merged', True)
            fwjrFile["merged"] = True

        wmbsFile = self.createFileFromDataStructsFile(fname=fwjrFile, jobID=jobID)

        if jobType == "Merge":
            self.wmbsMergeFilesToBuild.append(wmbsFile)
        else:
            self.wmbsFilesToBuild.append(wmbsFile)

        if fwjrFile["merged"]:
            self.addFileToDBS(fwjrFile, task,
                              jobType == "Repack" and fwjrFile["size"] > self.maxAllowedRepackOutputSize)

        return wmbsFile

    def handleJob(self, jobID, fwkJobReport):
        """
        _handleJob_

        Figure out if a job was successful or not, handle it appropriately
        (parse FWJR, update WMBS) and return the success status as a boolean

        """
        jobSuccess = fwkJobReport.taskSuccessful()
        logging.info("Task successful: %s", jobSuccess)

        outputMap = self.getOutputMapAction.execute(jobID=jobID,
                                                    conn=self.getDBConn(),
                                                    transaction=self.existingTransaction())

        jobType = self.getJobTypeAction.execute(jobID=jobID,
                                                conn=self.getDBConn(),
                                                transaction=self.existingTransaction())

        if jobSuccess:
            fileList = fwkJobReport.getAllFiles()

            # Consistency check comparing outputMap to fileList
            # they should match except for some limited special cases related to Tier-0:
            #     Repack Merge: May not produce Error output modules
            #     Express Merge: May not write RAW output
            # as of #7998, workflow output identifier is made of output module name and datatier
            outputModules = set([])
            for fwjrFile in fileList:
                outputModules.add(fwjrFile['outputModule'] + fwjrFile['dataset'].get('dataTier', ''))

            if set(outputMap) == outputModules:
                pass
            elif jobType == "LogCollect" and not outputMap and outputModules == {'LogCollect'}:
                pass
            elif jobType == "Merge" and (set(outputMap) & {'MergedErrorRAW', 'MergedErrorL1SCOUT', 'MergedErrorHLTSCOUT'}):
                pass
            elif jobType == "Express" and set(outputMap).difference(outputModules) == {'write_RAWRAW'}:
                pass
            else:
                # any job that is not multi-step and Processing/Production must FAIL!
                if jobType in ["Processing", "Production"]:
                    cmsRunSteps = sum([1 for step in fwkJobReport.listSteps() if step.startswith("cmsRun")])
                    if cmsRunSteps == 1:
                        jobSuccess = False
                        logging.info("Production/Processing single cmsRun step job does not match output modules.")
                    else:
                        msg = f"Job {jobID} accepted for multi-step CMSSW, even though "
                        msg += "the expected outputModules does not match content of the FWJR."
                        logging.warning(msg)
                else:
                    jobSuccess = False
                    logging.info("Non Production/Processing job does not match output modules.")

                if jobSuccess is False:
                    sortedOutputMap = sorted(outputMap.keys())
                    errMsg = f"Failing job {jobID}, with expected output modules: {sortedOutputMap}, "
                    errMsg += f"but has FWJR output modules: {sorted(outputModules)}"
                    logging.error(errMsg)
                    # override file list by the logArch1 output only
                    fileList = fwkJobReport.getAllFilesFromStep(step='logArch1')
        else:
            fileList = fwkJobReport.getAllFilesFromStep(step='logArch1')

        # Workaround: make sure every file has a valid location. See:
        # https://github.com/dmwm/WMCore/issues/9353 and https://github.com/dmwm/WMCore/issues/12092
        for fwjrFile in fileList:
            # T0 has analysis file without any location, see:
            # https://github.com/dmwm/WMCore/issues/9497
            if not fwjrFile.get("locations") and fwjrFile.get("lfn", "").endswith(".root"):
                logging.warning("The following file does not have any location: %s", fwjrFile)
                jobSuccess = False
                fileList = fwkJobReport.getAllFilesFromStep(step='logArch1')
                break

        if jobSuccess:
            logging.info("Job %d , handle successful job", jobID)
        else:
            logging.warning("Job %d , bad jobReport, failing job", jobID)

        # make sure the task name is present in FWJR (recover from WMBS if needed)
        if fileList:
            if jobSuccess:
                self.isTaskExistInFWJR(fwkJobReport, "success")
            else:
                self.isTaskExistInFWJR(fwkJobReport, "failed")

        # special check for LogCollect jobs
        skipLogCollect = False
        if jobSuccess and jobType == "LogCollect":
            for fwjrFile in fileList:
                try:
                    # this assumes there is only one file for LogCollect jobs, not sure what happend if that changes
                    self.associateLogCollectToParentJobsInWMStats(fwkJobReport, fwjrFile["lfn"],
                                                                  fwkJobReport.getTaskName())
                except Exception as ex:
                    skipLogCollect = True
                    logging.error("Error occurred: associating log collect location, will try again\n %s", str(ex))
                    break

        # now handle the job (unless the special LogCollect check failed)
        if not skipLogCollect:

            wmbsJob = Job(id=jobID)
            wmbsJob.load()
            outputID = wmbsJob.loadOutputID()
            wmbsJob.getMask()

            wmbsJob["fwjr"] = fwkJobReport

            if jobSuccess:
                wmbsJob["outcome"] = "success"
            else:
                wmbsJob["outcome"] = "failure"

            for fwjrFile in fileList:

                logging.debug("Job %d , register output %s", jobID, fwjrFile["lfn"])

                wmbsFile = self.addFileToWMBS(jobType, fwjrFile, wmbsJob["mask"],
                                              jobID=jobID, task=fwkJobReport.getTaskName())
                merged = fwjrFile['merged']
                moduleLabel = fwjrFile["module_label"]

                if merged:
                    self.mergedOutputFiles.append(wmbsFile)

                self.filesetAssoc.append({"lfn": wmbsFile["lfn"], "fileset": outputID})

                # LogCollect jobs have no output fileset
                if jobType == "LogCollect":
                    pass
                # Repack jobs that wrote too large merged output skip output filesets
                elif jobType == "Repack" and merged and wmbsFile["size"] > self.maxAllowedRepackOutputSize:
                    pass
                else:
                    dataTier = fwjrFile['dataset'].get('dataTier', '')
                    outputFilesets = self.outputFilesetsForJob(outputMap, merged, moduleLabel, dataTier)
                    for outputFileset in outputFilesets:
                        self.filesetAssoc.append({"lfn": wmbsFile["lfn"], "fileset": outputFileset})

            # Check if the job had any skipped files, put them in ACDC containers
            # We assume full file processing (no job masks)
            if jobSuccess:
                skippedFiles = fwkJobReport.getAllSkippedFiles()
                if skippedFiles and jobType not in ['LogCollect', 'Cleanup']:
                    self.jobsWithSkippedFiles[jobID] = skippedFiles

            if jobSuccess:
                self.listOfJobsToSave.append(wmbsJob)
            else:
                self.listOfJobsToFail.append(wmbsJob)

        return jobSuccess

    def associateLogCollectToParentJobsInWMStats(self, fwkJobReport, logAchiveLFN, task):
        """
        _associateLogCollectToParentJobsInWMStats_

        Associate a logArchive output to its parent job
        """
        if self.fwjrCouchDB is None:
            self.fwjrCouchDB = self.jobCouchdb.connectDatabase("%s/fwjrs" % self.jobDBName)

        inputFileList = fwkJobReport.getAllInputFiles()
        requestName = task.split('/')[1]
        keys = []
        for inputFile in inputFileList:
            keys.append([requestName, inputFile["lfn"]])
        resultRows = self.fwjrCouchDB.loadView("FWJRDump", 'jobsByOutputLFN',
                                               options={"stale": "update_after"},
                                               keys=keys)['rows']
        if resultRows:
            # get data from wmbs
            parentWMBSJobIDs = []
            for row in resultRows:
                parentWMBSJobIDs.append({"jobid": row["value"]})
            # update Job doc in wmstats
            results = self.getJobInfoByID.execute(parentWMBSJobIDs)
            parentJobNames = []

            if isinstance(results, list):
                for jobInfo in results:
                    parentJobNames.append(jobInfo['name'])
            else:
                parentJobNames.append(results['name'])

            self.localWMStats.updateLogArchiveLFN(parentJobNames, logAchiveLFN)
        else:
            # TODO: if the couch db is consistent with DB this should be removed (checking resultRow > 0)
            # It need to be failed and retried.
            logging.error(
                "job report is missing for updating log archive mapping\n Input file list\n %s", inputFileList)

        return

    def createMissingFWKJR(self, errorCode=999, errorDescription='Failure of unknown type'):
        """
        _createMissingFWJR_

        Create a missing FWJR if the report can't be found by the code in the
        path location.
        """
        report = Report()
        report.addError("cmsRun1", errorCode, "MissingJobReport", errorDescription)
        report.data.cmsRun1.status = "Failed"
        return report

    def createFilesInDBSBuffer(self):
        """
        _createFilesInDBSBuffer_
        It does the actual job of creating things in DBSBuffer
        WARNING: This assumes all files in a job have the same final location
        """
        if not self.dbsFilesToCreate:
            # Whoops, nothing to do!
            return

        dbsFileTuples = []
        dbsFileLoc = []
        dbsCksumBinds = []
        runLumiBinds = []
        jobLocations = set()

        for dbsFile in self.dbsFilesToCreate:
            # Append a tuple in the format specified by DBSBufferFiles.Add
            # Also run insertDatasetAlgo

            assocID = None
            datasetAlgoPath = '%s:%s:%s:%s:%s:%s:%s:%s' % (dbsFile['datasetPath'],
                                                           dbsFile["appName"],
                                                           dbsFile["appVer"],
                                                           dbsFile["appFam"],
                                                           dbsFile["psetHash"],
                                                           dbsFile['processingVer'],
                                                           dbsFile['acquisitionEra'],
                                                           dbsFile['globalTag'])
            # First, check if this is in the cache
            if datasetAlgoPath in self.datasetAlgoPaths:
                for da in self.datasetAlgoID:
                    if da['datasetAlgoPath'] == datasetAlgoPath:
                        assocID = da['assocID']
                        break

            if not assocID:
                # Then we have to get it ourselves
                try:
                    assocID = dbsFile.insertDatasetAlgo()
                    self.datasetAlgoPaths.append(datasetAlgoPath)
                    self.datasetAlgoID.append({'datasetAlgoPath': datasetAlgoPath,
                                               'assocID': assocID})
                except WMException:
                    raise
                except Exception as ex:
                    msg = "Unhandled exception while inserting datasetAlgo: %s\n" % datasetAlgoPath
                    msg += str(ex)
                    logging.error(msg)
                    raise AccountantWorkerException(msg)

            # Associate the workflow to the file using the taskPath and the requestName
            # TODO: debug why it happens and then drop/recover these cases automatically
            taskPath = dbsFile.get('task')
            if not taskPath:
                msg = "Can't do workflow association, report this error to a developer.\n"
                msg += "DbsFile : %s" % str(dbsFile)
                raise AccountantWorkerException(msg)
            workflowName = taskPath.split('/')[1]
            workflowPath = '%s:%s' % (workflowName, taskPath)
            if workflowPath in self.workflowPaths:
                for wf in self.workflowIDs:
                    if wf['workflowPath'] == workflowPath:
                        workflowID = wf['workflowID']
                        break
            else:
                result = self.dbsGetWorkflow.execute(workflowName, taskPath, conn=self.getDBConn(),
                                                     transaction=self.existingTransaction())
                workflowID = result['id']

            self.workflowPaths.append(workflowPath)
            self.workflowIDs.append({'workflowPath': workflowPath, 'workflowID': workflowID})

            lfn = dbsFile['lfn']
            selfChecksums = dbsFile['checksums']
            jobLocation = dbsFile.getLocations()[0]
            jobLocations.add(jobLocation)
            dbsFileTuples.append((lfn, dbsFile['size'],
                                  dbsFile['events'], assocID,
                                  dbsFile['status'], workflowID, dbsFile['in_phedex']))

            dbsFileLoc.append({'lfn': lfn, 'pnn': jobLocation})
            if dbsFile['runs']:
                runLumiBinds.append({'lfn': lfn, 'runs': dbsFile['runs']})

            if selfChecksums:
                # If we have checksums we have to create a bind
                # For each different checksum
                for entry in selfChecksums:
                    dbsCksumBinds.append({'lfn': lfn, 'cksum': selfChecksums[entry],
                                          'cktype': entry})

        try:

            diffLocation = jobLocations.difference(self.dbsLocations)
            self.dbsInsertLocation.execute(siteName=diffLocation,
                                           conn=self.getDBConn(),
                                           transaction=self.existingTransaction())
            self.dbsLocations.union(diffLocation)  # update the component cache location list

            self.dbsCreateFiles.execute(files=dbsFileTuples,
                                        conn=self.getDBConn(),
                                        transaction=self.existingTransaction())

            self.dbsSetLocation.execute(binds=dbsFileLoc,
                                        conn=self.getDBConn(),
                                        transaction=self.existingTransaction())

            self.dbsSetChecksum.execute(bulkList=dbsCksumBinds,
                                        conn=self.getDBConn(),
                                        transaction=self.existingTransaction())

            if runLumiBinds:
                self.dbsSetRunLumi.execute(file=runLumiBinds,
                                           conn=self.getDBConn(),
                                           transaction=self.existingTransaction())
        except WMException:
            raise
        except Exception as ex:
            msg = "Got exception while inserting files into DBSBuffer!\n"
            msg += str(ex)
            logging.error(msg)
            logging.debug("Listing binds:")
            logging.debug("jobLocation: %s", jobLocation)
            logging.debug("dbsFiles: %s", dbsFileTuples)
            logging.debug("dbsFileLoc: %s", dbsFileLoc)
            logging.debug("Checksum binds: %s", dbsCksumBinds)
            logging.debug("RunLumi binds: %s", runLumiBinds)
            raise AccountantWorkerException(msg)

        # Now that we've created those files, clear the list
        self.dbsFilesToCreate = []
        return

    def handleWMBSFiles(self, wmbsFilesToBuild, parentageBinds):
        """
        _handleWMBSFiles_

        Do what can be done in bulk in bulk
        """
        if not wmbsFilesToBuild:
            # Nothing to do
            return

        runLumiBinds = []
        fileCksumBinds = []
        fileLocations = []
        fileCreate = []

        for wmbsFile in wmbsFilesToBuild:
            lfn = wmbsFile['lfn']
            if lfn is None:
                continue

            selfChecksums = wmbsFile['checksums']
            # by jobType add to different parentage relation
            # if it is the merge job, don't include the parentage on failed input files.
            # otherwise parentage is set for all input files.
            parentageBinds.append({'child': lfn, 'jobid': wmbsFile['jid']})

            if wmbsFile['runs']:
                runLumiBinds.append({'lfn': lfn, 'runs': wmbsFile['runs']})

            if wmbsFile.getLocations():
                outpnn = wmbsFile.getLocations()[0]
                if self.pnn2Psn.get(outpnn, None):
                    fileLocations.append({'lfn': lfn, 'location': outpnn})
                else:
                    msg = "PNN doesn't exist in wmbs_pnns table: %s (investigate)" % outpnn
                    logging.error(msg)
                    raise AccountantWorkerException(msg)

            if selfChecksums:
                # If we have checksums we have to create a bind
                # For each different checksum
                for entry in selfChecksums:
                    fileCksumBinds.append({'lfn': lfn, 'cksum': selfChecksums[entry],
                                           'cktype': entry})

            fileCreate.append([lfn,
                               wmbsFile['size'],
                               wmbsFile['events'],
                               None,
                               wmbsFile["first_event"],
                               wmbsFile['merged']])

        if not fileCreate:
            return

        try:

            self.addFileAction.execute(files=fileCreate,
                                       conn=self.getDBConn(),
                                       transaction=self.existingTransaction())

            if runLumiBinds:
                self.setFileRunLumi.execute(file=runLumiBinds,
                                            conn=self.getDBConn(),
                                            transaction=self.existingTransaction())

            if fileCksumBinds:
                self.setFileAddChecksum.execute(bulkList=fileCksumBinds,
                                                conn=self.getDBConn(),
                                                transaction=self.existingTransaction())

            self.setFileLocation.execute(lfn=fileLocations,
                                         conn=self.getDBConn(),
                                         transaction=self.existingTransaction())


        except WMException:
            raise
        except Exception as ex:
            msg = "Error while adding files to WMBS!\n"
            msg += str(ex)
            logging.error(msg)
            logging.debug("Printing binds:")
            logging.debug("FileCreate binds: %s", fileCreate)
            logging.debug("Runlumi binds: %s", runLumiBinds)
            logging.debug("Checksum binds: %s", fileCksumBinds)
            logging.debug("FileLocation binds: %s", fileLocations)
            raise AccountantWorkerException(msg)

        # Clear out finished files
        wmbsFilesToBuild = []
        return

    def createFileFromDataStructsFile(self, fname, jobID):
        """
        _createFileFromDataStructsFile_

        This function will create a WMBS File given a DataStructs file
        """
        wmbsFile = File()
        wmbsFile.update(fname)

        if fname["locations"] and isinstance(fname["locations"], set):
            pnn = list(fname["locations"])[0]
        elif fname["locations"] and isinstance(fname["locations"], list):
            if len(fname['locations']) > 1:
                logging.error("Have more then one location for a file in job %i", jobID)
                logging.error("Choosing location %s", fname['locations'][0])
            pnn = fname["locations"][0]
        else:
            pnn = fname["locations"]

        wmbsFile["locations"] = set()

        if pnn != None:
            wmbsFile.setLocation(pnn=pnn, immediateSave=False)
        wmbsFile['jid'] = jobID

        return wmbsFile

    def handleDBSBufferParentage(self):
        """
        _handleDBSBufferParentage_

        Handle all the DBSBuffer Parentage in bulk if you can
        """
        outputLFNs = [f['lfn'] for f in self.mergedOutputFiles]
        bindList = []
        for lfn in outputLFNs:
            newParents = self.findDBSParents(lfn=lfn)
            for parentLFN in newParents:
                bindList.append({'child': lfn, 'parent': parentLFN})

        # Now all the parents should exist
        # Commit them to DBSBuffer
        logging.info("About to commit DBSBuffer heritage information for: %d binds", len(bindList))

        if bindList:
            try:
                self.dbsLFNHeritage.execute(binds=bindList,
                                            conn=self.getDBConn(),
                                            transaction=self.existingTransaction())
            except WMException:
                raise
            except Exception as ex:
                msg = "Error while trying to handle the DBS LFN heritage\n"
                msg += str(ex)
                msg += "BindList: %s" % bindList
                logging.error(msg)
                raise AccountantWorkerException(msg)
        return

    def handleSkippedFiles(self):
        """
        _handleSkippedFiles_

        Handle all the skipped files in bulk,
        the way it handles the skipped files
        imposes an important restriction:
        Skipped files should have been processed by a single job
        in the task and no job mask exists in it.
        This is suitable for jobs using ParentlessMergeBySize/FileBased/MinFileBased
        splitting algorithms.
        Here ACDC records and created and the file are moved
        to wmbs_sub_files_failed from completed.
        """
        jobList = self.getFullJobInfo.execute([{'jobid': x} for x in self.jobsWithSkippedFiles],
                                              fileSelection=self.jobsWithSkippedFiles,
                                              conn=self.getDBConn(),
                                              transaction=self.existingTransaction())
        self.dataCollection.failedJobs(jobList, useMask=False)
        return



#!/usr/bin/env python
"""
_MarkFinishedSubscriptions_

MySQL implementation of Subscriptions.MarkFinishedSubscriptions

Created on Aug 29, 2012

@author: dballest
"""

from time import time

from WMCore.Database.DBFormatter import DBFormatter

class MarkFinishedSubscriptions(DBFormatter):
    """
    Marks the given subscriptions as finished, and updates the timestamp
    """

    updateSQL = """UPDATE wmbs_subscription
                   SET finished = :finished, last_update = :timestamp
                   WHERE id = :id"""

    def execute(self, ids, finished = True, conn = None,
                transaction = False):
        """
        _execute_

        Update the subscriptions to match their finished status
        """

        if finished:
            finished = 1
        else:
            finished = 0

        #Make sure it's a list of IDs
        if not isinstance(ids, list):
            ids = [ids]

        binds = []
        for subId in ids:
            binds.append({'id': subId, 'finished': finished, 'timestamp': int(time())})

        if binds:
            self.dbi.processData(self.updateSQL, binds, conn = conn,
                                 transaction = transaction)

        return
